#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Parsing utilities for extracting dependencies from SQL Alchemy select statements."""
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

from sqlalchemy.orm.util import _ORMJoin
from sqlalchemy.schema import Table as SATable
from sqlalchemy.sql.elements import (
    BinaryExpression,
    BindParameter,
    ClauseList,
    ColumnClause,
    ColumnElement,
    TextClause,
)
from sqlalchemy.sql.expression import Alias, BooleanClauseList, Function, Label  # type: ignore
from sqlalchemy.sql.expression import Select as _Select
from sqlalchemy.sql.schema import Column as SAColumn

from dbgen.exceptions import QueryParsingError
from dbgen.utils.lists import flatten

# TODO Allow for arrow expressions


def expand_clause_list(clause_statement):
    output = []
    if getattr(clause_statement, "clauses", None) is not None:
        for element in getattr(clause_statement, "clauses"):
            if getattr(clause_statement, "clauses", None) is not None:
                output.extend(expand_clause_list(element))
            else:
                output.append(element)
    else:
        return [clause_statement]
    return output


def get_where_dependency(where_stmt: BooleanClauseList):
    tables = []
    columns = []
    fks = []
    for child in where_stmt.get_children():
        if isinstance(child, SATable):
            tables.append(child)
        elif isinstance(child, SAColumn) and child.foreign_keys:
            fks.append(child)
        elif isinstance(child, SAColumn):
            columns.append(expand_col(child))
        elif isinstance(child, (_ORMJoin, ColumnElement)):
            c_tables, c_columns, c_fks = get_from_dependency(child)
            tables.extend(c_tables)
            columns.extend(c_columns)
            fks.extend(c_fks)
        else:
            print(type(child))
    return columns, tables, fks


def expand_col(column: ColumnElement):
    if isinstance(column, SAColumn):
        if isinstance(column.table, Alias):
            if isinstance(column.table.original, SATable):
                return [column.table.original.columns.get(column.name)]
            else:
                raise NotImplementedError(column.table.original, type(column.table.original))
        return [column]
    elif isinstance(column, Label):
        return expand_col(column.element)
    elif isinstance(column, BinaryExpression):
        return [*expand_col(column.left), *expand_col(column.right)]
    elif isinstance(column, (BindParameter, TextClause)):
        return []
    elif getattr(column, 'get_children', None):
        return list(flatten(map(expand_col, column.get_children())))
    else:
        raise NotImplementedError(f"Unknown selected column type:\n{column}\n{type(column)}")


def get_select_dependency(select_stmt: _Select):
    selected = set(select_stmt.selected_columns)
    columns = set()
    for col in selected:
        columns.update(expand_col(col))

    if select_stmt._having_criteria is not None:  # type: ignore
        for criteria in select_stmt._having_criteria:  # type: ignore
            columns.update(expand_col(criteria))

    tables = {column.table for column in columns}
    fks = {column for column in columns if column.foreign_keys}

    return columns, tables, fks


def get_from_dependency(from_statement: Union[_ORMJoin, ColumnElement, ClauseList]):
    tables = set()
    columns = set()
    fks = set()
    for child in from_statement.get_children():
        if isinstance(child, Alias):
            tables.add(child.original)
        elif isinstance(child, SATable):
            tables.add(child)
        elif isinstance(child, SAColumn) and child.foreign_keys:
            fks.add(child)
        elif isinstance(child, SAColumn):
            columns.update(expand_col(child))
        elif isinstance(child, (_ORMJoin, ColumnElement, ClauseList)):
            c_columns, c_tables, c_fks = get_from_dependency(child)
            tables.update(c_tables)
            columns.update(c_columns)
            fks.update(c_fks)
        else:
            print(type(child))
    return columns, tables, fks


def get_statement_dependency(
    select_stmt: _Select,
) -> Tuple[Set, Set, Set]:
    """Parses a sqlalchemy select statment to get its dependencies.

    Parses the select, where, order_by, and group by clause for all properties, tables, and foreign keys
    required to run the query.

    Args:
        select_stmt (_Select): A SQL alchemy select statement

    Returns:
        Tuple[Set[str],Set[str],Set[str]]: 3 sets of the column, table, and fk depencies
    """
    # Parse the select clause
    select_cols, select_tabs, select_fks = get_select_dependency(select_stmt)
    # Parse the where clause
    if select_stmt.whereclause is not None:
        where_cols, where_tabs, where_fks = get_from_dependency(select_stmt.whereclause)
    else:
        where_cols, where_tabs, where_fks = set(), set(), set()

    from_cols, from_fks, from_tabs = set(), set(), set()
    # compile the from statements
    from_statements = select_stmt.get_final_froms()  # type: ignore
    if len(from_statements) > 1:
        print("Warning: 2 from statements detected possible crossjoin used")
    # Add the order_by clause if set
    if select_stmt._order_by_clause is not None:  # type: ignore
        from_statements.append(select_stmt._order_by_clause)  # type: ignore
    # Add the group_by clause if set
    if select_stmt._group_by_clause is not None:  # type: ignore
        from_statements.append(select_stmt._group_by_clause)  # type: ignore

    for from_statement in from_statements:
        c_from_cols, c_from_tabs, c_from_fks = get_from_dependency(from_statement)
        from_cols.update(c_from_cols)
        from_tabs.update(c_from_tabs)
        from_fks.update(c_from_fks)

    # merge dependencies
    cols = select_cols | where_cols | from_cols
    tabs = select_tabs | where_tabs | from_tabs
    fks = select_fks | where_fks | from_fks
    return cols, tabs, fks


def _get_select_keys(select_statement: _Select) -> List[str]:
    output_keys: Dict[str, Optional[str]] = {}

    for column in select_statement.selected_columns:
        col_key, marker = _parse_column(column)
        if col_key in output_keys:
            raise ValueError(
                f"Ambiguous Column Name: Both {marker} and {output_keys[col_key]} have column {col_key}.\n"
                "Please alias one of them using something like col.label('alias')"
            )
        output_keys[col_key] = marker
    return list(output_keys.keys())


def _parse_column(column: Any) -> Tuple[str, Optional[str]]:

    try:
        if isinstance(column, SAColumn):
            col_key = column.name
            marker = f"{column.table.name}.{col_key}"
            return (col_key, marker)
        elif isinstance(column, Alias):
            col_key = column.name  # type: ignore
            marker = f"{column.table.name}.{col_key}"  # type: ignore
            return (col_key, marker)
        elif isinstance(column, Label):
            col_key = column.key
            expanded_col = expand_col(column)
            marker = expanded_col
            return (col_key, marker)
        elif isinstance(column, ColumnClause):
            expanded_col = expand_col(column)
            if len(expanded_col) == 0:
                if isinstance(str, column.name):
                    column.name = cast(str, column.name)  # type: ignore
                    return (column.name, None)
                raise QueryParsingError(f"Unknown column name: {column}")
        elif isinstance(column, (Function, BinaryExpression)):
            raise QueryParsingError(
                "SQLAlchemy Functions need to be labeld due to the imprecise naming of function columns in sqlalchemy.\n"
                f"Try something like \"{getattr(column,'name','COLUMN_NAME')}(YOUR_COLUMN).label('my_column_name')\""
            )
    except QueryParsingError:
        raise
    except AttributeError as exc:
        raise QueryParsingError(
            "Error occurred during parsing that led to an internal sqlalchemy error\n" f"Column: {column}"
        ) from exc

    raise NotImplementedError(str(column), type(column))



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Objects related to the running of Models and Generators."""
from bdb import BdbQuit
from datetime import datetime, timedelta
from math import ceil
from time import time
from typing import Any, Dict, List, Mapping, Optional, Set, Tuple
from uuid import UUID

from psycopg import connect as pg3_connect
from pydantic.fields import Field, PrivateAttr
from pydasher import hasher
from sqlalchemy.future import Engine
from sqlmodel import Session, select
from tqdm import tqdm

import dbgen.exceptions as exceptions
from dbgen.core.base import Base, encoders
from dbgen.core.generator import Generator
from dbgen.core.metadata import (
    GeneratorEntity,
    GeneratorRunEntity,
    GensToRun,
    ModelEntity,
    Repeats,
    RunEntity,
    Status,
)
from dbgen.core.model import Model
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import BaseQuery
from dbgen.exceptions import (
    DBgenExternalError,
    DBgenSkipException,
    RepeatException,
    SerializationError,
    ValidationError,
)
from dbgen.utils.log import LogLevel


class RunConfig(Base):
    """Configuration for the running of a Generator and Model"""

    retry: bool = False
    include: Set[str] = Field(default_factory=set)
    exclude: Set[str] = Field(default_factory=set)
    start: Optional[str]
    until: Optional[str]
    batch_size: Optional[int]
    progress_bar: bool = True
    skip_row_count: bool = False
    skip_on_error: bool = False
    batch_number: int = 10
    log_level: LogLevel = LogLevel.INFO

    def should_gen_run(self, generator: Generator) -> bool:
        """Check a generator against include/exclude to see if it should run."""
        markers = [generator.name, *generator.tags]
        should_run = any(
            map(
                lambda x: (not self.include or x in self.include) and x not in self.exclude,
                markers,
            )
        )
        return should_run

    def get_invalid_markers(self, model: Model) -> Dict[str, List[str]]:
        """Check that all inputs to RunConfig are meaningful for the model."""
        invalid_marks = {}
        gen_names = model.gens().keys()
        # Validate start and until
        for attr in ("start", "until"):
            val: str = getattr(self, attr)
            if val is not None and val not in gen_names:
                invalid_marks[attr] = [val]
        # Validate include and exclude as sets
        for attr in ("include", "exclude"):
            set_val: Set[str] = getattr(self, attr)
            invalid_vals = [x for x in set_val if not model.validate_marker(x)]
            if invalid_vals:
                invalid_marks[attr] = invalid_vals

        return invalid_marks


def update_run_by_id(run_id, status: Status, session: Session):
    run = session.get(RunEntity, run_id)
    assert run, f"No run found with id {run_id}"
    run.status = status
    session.commit()


class RunInitializer(Base):
    """Intializes a run by syncing the database and getting the run_id."""

    def execute(self, engine: Engine, run_config: RunConfig) -> int:
        # Use some metadatabase connection to initialize initialize the run
        # Store the details of the run on the metadatabase so downstream GeneratorRuns can pick them up
        # Sync the database with the registries
        with Session(engine) as session:
            run = RunEntity(status=Status.initialized)
            session.add(run)
            session.commit()
            session.refresh(run)
            assert isinstance(run.id, int)
            run.status = Status.running
            session.commit()
            run_id = run.id

        return run_id


class BaseGeneratorRun(Base):
    """A lightwieght wrapper for the Generator that grabs a specific Generator from metadatabase and runs it."""

    _old_repeats: Set[UUID] = PrivateAttr(default_factory=set)
    _new_repeats: Set[UUID] = PrivateAttr(default_factory=set)

    def get_gen(self, meta_engine: Engine, *args, **kwargs) -> Generator:
        raise NotImplementedError

    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_id: Optional[int],
        run_config: Optional[RunConfig],
        ordering: Optional[int],
    ):

        # Set default values for run_config if none provided
        if run_config is None:
            run_config = RunConfig()

        generator = self.get_gen(meta_engine=meta_engine)
        # Initialize the generator_row in the meta database
        meta_session = Session(meta_engine)

        gen_run = self._initialize_gen_run(
            generator=generator, session=meta_session, run_id=run_id, ordering=ordering
        )
        # Check if our run config excludes our generator
        if not run_config.should_gen_run(generator):
            self._logger.info(f'Excluding generator {generator.name!r}')
            gen_run.status = Status.excluded
            meta_session.commit()
            return
        # Start the Generator
        self._logger.info(f'Running generator {generator.name!r}...')
        gen_run.status = Status.running
        meta_session.commit()
        start = time()
        # Set the extractor
        self._logger.debug('Initializing extractor')
        extractor_connection = main_engine.connect()
        extract = generator.extract
        if isinstance(extract, BaseQuery):
            extract.set_extractor(connection=extractor_connection, yield_per=generator.batch_size)
        else:
            extract.set_extractor()
        self._logger.debug('Fetching extractor length')
        if not run_config.skip_row_count:
            try:
                row_count = extract.length(connection=extractor_connection)
            except TypeError:
                self._logger.error(
                    f'Extract {extract}\'s length method does not accept the required kwargs, please add **_ kwarg to suppress this error.'
                )
                row_count = None
        else:
            row_count = None

        gen_run.inputs_extracted = row_count
        meta_session.commit()

        self._logger.debug('Fetching repeats')
        # Query the repeats table for input_hashes that match this generator's hash
        self._old_repeats = set(
            meta_session.exec(select(Repeats.input_hash).where(Repeats.generator_id == generator.uuid)).all()
        )
        # The batch_size is set either on the run_config or the generator
        batch_size = run_config.batch_size or generator.batch_size
        if batch_size is None and row_count:
            batch_size = ceil(row_count / run_config.batch_number)
        assert batch_size is None or batch_size > 0, f"Invalid batch size batch_size must be >0: {batch_size}"
        # Open raw connections for fast loading
        main_raw_connection = pg3_connect(str(main_engine.url))
        meta_raw_connection = meta_engine.raw_connection()
        batch_done = lambda x: (x and x % batch_size == 0) if batch_size is not None else False
        # Start while loop to iterate through the nodes
        self._logger.info('Looping through extracted rows...')
        progress_bar = tqdm(
            total=row_count,
            position=1,
            leave=False,
            unit=" Rows",
            disable=not run_config.progress_bar,
        )
        batch_num = 1
        total_batches = f" of {str(ceil(row_count / batch_size))}" if row_count and batch_size else ''
        batch_message = "Batch {}{}: {:10}"
        progress_bar.set_description(batch_message.format(batch_num, total_batches, 'Transforming'))
        try:
            while True:
                row: Dict[str, Mapping[str, Any]] = {}
                try:
                    for node in generator._sort_graph():
                        output = node.run(row)
                        # Extract outputs need to be fed to our repeat checker and need to be checked for stop iterations
                        if isinstance(node, Extract):
                            if output is None or batch_done(gen_run.inputs_processed):
                                load_msg = batch_message.format(batch_num, total_batches, 'Loading')
                                progress_bar.set_description(load_msg)
                                self._logger.debug(load_msg)
                                self._load_repeats(meta_raw_connection, generator)
                                rows_inserted, rows_updated = self._load(main_raw_connection, generator)
                                gen_run.rows_inserted += rows_inserted
                                gen_run.rows_updated += rows_updated
                                meta_session.commit()
                                self._logger.debug('done loading batch.')
                                self._logger.debug(f'inserted {rows_inserted} rows.')
                                self._logger.debug(f'updated {rows_updated} rows.')
                                # Increment batch and move to T-step
                                batch_num += 1
                                transform_msg = batch_message.format(batch_num, total_batches, 'Transforming')
                                progress_bar.set_description(transform_msg)
                                self._logger.debug(transform_msg)
                            # if we are out of rows break out of while loop
                            if output is None:
                                raise StopIteration
                            gen_run.inputs_processed += 1
                            is_repeat, input_hash = self._check_repeat(output, generator.uuid)
                            if not run_config.retry and is_repeat:
                                raise RepeatException()
                        row[node.hash] = output  # type: ignore
                    if not is_repeat:
                        self._new_repeats.add(input_hash)
                        gen_run.unique_inputs += 1
                    progress_bar.update()
                # Stop iteration is used to catch the empty extractor
                except StopIteration:
                    break
                # A repeated input from the extract will also cause a row to be skipped
                except RepeatException:
                    progress_bar.update()
                    continue
                # Any node can raise a skip exception to skip the input before loading
                except DBgenSkipException as exc:
                    self._logger.debug(f"Skipped Row: {exc.msg}")
                    gen_run.inputs_skipped += 1
                    progress_bar.update()
                # External errors are raised whenever a node fails due to internal logic
                except (DBgenExternalError, ValidationError) as e:
                    msg = f"Error when running generator {generator.name}"
                    self._logger.error(msg)
                    self._logger.error(e)
                    if run_config.skip_on_error:
                        gen_run.inputs_skipped += 1
                        progress_bar.update()
                        continue
                    # self._logger.error(f"\n{e}")
                    gen_run.status = Status.failed
                    gen_run.error = str(e)
                    run = meta_session.get(RunEntity, run_id)
                    assert run
                    run.errors = run.errors + 1 if run.errors else 1
                    meta_session.commit()
                    meta_session.close()
                    return 2
                except (
                    Exception,
                    KeyboardInterrupt,
                    SystemExit,
                    BdbQuit,
                ) as e:
                    gen_run.status = Status.failed
                    gen_run.error = (
                        f"Uncaught Error encountered during running of generator {generator.name}: {e!r}"
                    )
                    update_run_by_id(run_id, Status.failed, meta_session)
                    raise
                # finally:
                #     progress_bar.update()
        # Close all connections
        finally:
            gen_run.runtime = round(time() - start, 3)
            meta_session.commit()
            main_raw_connection.close()
            meta_raw_connection.close()
            extractor_connection.close()

        gen_run.status = Status.completed
        gen_run.runtime = round(time() - start, 3)
        self._logger.info(
            f"Finished running generator {generator.name}({generator.uuid}) in {gen_run.runtime}(s)."
        )
        self._logger.info(f"Loaded approximately {gen_run.rows_inserted} rows")
        meta_session.commit()
        meta_session.close()
        return 0

    def _initialize_gen_run(
        self,
        session: Session,
        generator: Generator,
        run_id: Optional[int],
        ordering: Optional[int],
    ) -> GeneratorRunEntity:
        # if no run_id is provided create one and mark it as a testing run
        if run_id is None:
            run = RunEntity(status='testing')
            session.add(run)
            session.commit()
            session.refresh(run)
            ordering = 0
            run_id = run.id
        gen_row = generator._get_gen_row()
        session.merge(gen_row)
        session.commit()
        query = generator.extract.query if isinstance(generator.extract, BaseQuery) else ''
        gen_run = GeneratorRunEntity(
            run_id=run_id,
            generator_id=gen_row.id,
            status=Status.initialized,
            ordering=ordering,
            query=query,
        )
        session.add(gen_run)
        session.commit()
        session.refresh(gen_run)
        return gen_run

    def _load(self, connection, generator: Generator) -> Tuple[int, int]:
        rows_inserted = 0
        rows_updated = 0
        for load in generator._sorted_loads():
            if load.insert:
                rows_inserted += len(load._output)
            else:
                rows_updated += len(load._output)
            load.load(connection, gen_id=self.uuid)
        return (rows_inserted, rows_updated)

    def _load_repeats(self, connection, generator: Generator) -> None:
        rows = ((generator.uuid, input_hash) for input_hash in self._new_repeats)
        Repeats._quick_load(connection, rows, column_names=["generator_id", "input_hash"])
        self._old_repeats = self._old_repeats.union(self._new_repeats)
        self._new_repeats = set()

    def _check_repeat(self, extracted_dict: Dict[str, Any], generator_uuid: UUID) -> Tuple[bool, UUID]:
        # Convert Row to a dictionary so we can hash it for repeat-checking
        input_hash = UUID(hasher((generator_uuid, extracted_dict), encoders=encoders))
        # If the input_hash has been seen and we don't have retry=True skip row
        is_repeat = input_hash in self._old_repeats or input_hash in self._new_repeats
        return (is_repeat, input_hash)


class GeneratorRun(BaseGeneratorRun):
    generator: Generator

    def get_gen(self, meta_engine: Engine, *args, **kwargs):
        return self.generator


class RemoteGeneratorRun(BaseGeneratorRun):
    generator_id: UUID

    def get_gen(self, meta_engine, *args, **kwargs):
        with Session(meta_engine) as sess:
            gen_json = sess.exec(
                select(GeneratorEntity.gen_json).where(GeneratorEntity.id == self.generator_id)
            ).one()
            try:
                generator = Generator.deserialize(gen_json)
            except ModuleNotFoundError as exc:
                import os

                raise SerializationError(
                    f"While deserializing generator id {self.generator_id} an unknown module was encountered. Are you using custom dbgen objects reachable by your python environment? Make sure any custom extractors or code can be found in your PYTHONPATH environment variable\nError: {exc}\nPYTHONPATH={os.environ.get('PYTHONPATH')}"
                ) from exc
            if generator.uuid != self.generator_id:
                error = f"Deserialization Failed the generator hash has changed for generator named {generator.name}!\n{generator}\n{self.generator_id}"
                raise exceptions.SerializationError(error)
        return generator


class ModelRun(Base):
    model: Model

    def get_gen_run(self, generator: Generator) -> BaseGeneratorRun:
        return GeneratorRun(generator=generator)

    def execute(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_config: RunConfig = None,
        nuke: bool = False,
        rerun_failed: bool = False,
    ) -> RunEntity:
        start = time()
        if run_config is None:
            run_config = RunConfig()
        # Sync the Database statew with the model state
        self.model.sync(main_engine, meta_engine, nuke=nuke)

        # If doing last failed run query for gens to run and add to include
        if rerun_failed:
            with meta_engine.connect() as conn:
                result = conn.execute(select(GensToRun.__table__.c.name))
                for (gen_name,) in result:
                    run_config.include.add(gen_name)

        # Initialize the run
        run_init = RunInitializer()
        run_id = run_init.execute(meta_engine, run_config)
        sorted_generators = self.model._sort_graph()
        # Add generators to metadb
        with Session(meta_engine) as meta_session:
            model_row = self.model._get_model_row()
            model_row.last_run = datetime.now()
            existing_model = meta_session.get(ModelEntity, model_row.id)
            if not existing_model:
                meta_session.merge(model_row)
            else:
                existing_model.last_run = datetime.now()
            meta_session.commit()

        # Apply start and until to exclude generators not between start_idx and until_idx
        if run_config.start or run_config.until:
            gen_names = [gen.name for gen in sorted_generators]
            start_idx = gen_names.index(run_config.start) if run_config.start else 0
            until_idx = gen_names.index(run_config.until) + 1 if run_config.until else len(gen_names)
            # Modify include to only include the gen_names that pass the test
            run_config.include = run_config.include.union(gen_names[start_idx:until_idx])
            print(f"Only running generators: {gen_names[start_idx:until_idx]} due to start/until")
            self._logger.debug(
                f"Only running generators: {gen_names[start_idx:until_idx]} due to start/until"
            )
        with tqdm(
            total=len(sorted_generators),
            position=0,
            disable=not run_config.progress_bar,
            unit=' Generator',
        ) as tq:
            for i, generator in enumerate(sorted_generators):
                tq.set_description(generator.name)
                gen_run = self.get_gen_run(generator)
                gen_run.execute(main_engine, meta_engine, run_id, run_config, ordering=i)
                tq.update()

        # Complete run
        with Session(meta_engine) as session:
            update_run_by_id(run_id, Status.completed, session)
            run = session.get(RunEntity, run_id)
            assert run
            run.runtime = timedelta(seconds=time() - start)
            session.commit()
            session.refresh(run)
        return run


class RemoteModelRun(ModelRun):
    def get_gen_run(self, generator):
        return RemoteGeneratorRun(generator_id=generator.uuid)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typer import Argument, Option

from dbgen.cli.utils import version_callback

model_string_option = Option(None, '--model', envvar=["DBGEN_MODEL_STR", "dbgen_model_str"])
model_arg_option = Argument('--model', envvar=["DBGEN_MODEL_STR", "dbgen_model_str"])
version_option = Option(None, "--version", callback=version_callback, is_eager=True)
config_option = Option('.env', "--config", "-c", help="Configuration file.", envvar='DBGEN_CONFIG')



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from pathlib import Path
from types import FunctionType
from typing import TYPE_CHECKING

import typer

from dbgen.cli.styles import LOGO_STYLE, bad_typer_print
from dbgen.core.model import Model

if TYPE_CHECKING:
    from dbgen.utils.sql import Connection  # pragma: no cover

# Errors
ERROR_FORMAT = "Model is not in MODULE:PACKAGE format: {0}"
ERROR_MODULE = "Could not find module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_PACKAGE = "Could not find package within module:\nModule: {0}\nPackage: {1}\nError: {2}"
ERROR_NOT_MODEL = "Import String is not for a DBgen Model: \nImport String: {0}\nClass: {1}"
ERROR_NOT_MODEL_FUNCTION = "Import String is for a function that does not produce a DBgen Model: \nImport String: {0}\nOutput Class: {1}"
ERROR_RUNNING_MODEL_FACT = "Import String is for a function produced an error or required arguments: \nImport String: {0}\nOutput Class: {1} \n{2}{3}"

logger = logging.getLogger(__name__)

state = {"confirm": True}


# Callback functions for parsing and validating args
def confirm_nuke(value: bool):
    """
    Confirm that user wants to nuke database

    Args:
        value (bool): nuke_value
    """
    if value and state["confirm"]:
        confirm = input("Are you sure you want to nuke the database? (Y/n) ")
        if confirm.lower() != "y":
            exit(1)
    return value


def set_confirm(value: bool):
    """Auto confirm all prompted values"""
    state["confirm"] = not value


def file_existence(value: Path):
    if value and value.exists():
        return value
    elif value is not None:
        # No config found warn user
        logger.warning(f"Provided config {value.absolute()} does not exist.")
    return None


def version_callback(value: bool):
    """
    Eagerly print the version LOGO

    Args:
        value (bool): [description]

    Raises:
        typer.Exit: exits after showing version
    """
    if value:
        typer.echo(LOGO_STYLE)
        raise typer.Exit()


def validate_model_str(model_str: str) -> Model:
    """
    Validate the user input model import str checking for malformed and invalid inputs.py

    Args:
        model_str (str): CLI/config file input in MODULE:PACKAGE format

    Raises:
        typer.BadParameter: Whenever one of many checks is failed

    Returns:
        Model: the output model after checks are in place
    """
    basic_error = lambda fmt, val: typer.BadParameter(fmt.format(*val))

    if model_str is None:
        raise typer.BadParameter("--model is required.")

    # Check for empty string model_strs
    if model_str in (":", ""):
        raise basic_error(ERROR_FORMAT, [model_str])

    split_model = model_str.split(":")
    if len(split_model) != 2:
        raise basic_error(ERROR_FORMAT, [model_str])
    module, package = split_model
    try:
        _temp = __import__(module, globals(), locals(), [package])
        model = getattr(_temp, package)
        if isinstance(model, Model):
            return model
        elif isinstance(model, FunctionType):
            try:
                model = model()
            except TypeError as exc:
                print(exc)
                import traceback

                exc_str = traceback.format_exc()
                raise basic_error(
                    ERROR_RUNNING_MODEL_FACT, [model_str, type(model).__name__, "#" * 24 + "\n", str(exc_str)]
                ) from exc
            if isinstance(model, Model):
                return model
            raise basic_error(ERROR_NOT_MODEL_FUNCTION, [model_str, type(model).__name__])

        raise basic_error(ERROR_NOT_MODEL, [model_str, type(model).__name__])
    except ModuleNotFoundError as exc:
        if "No module" in str(exc):
            raise basic_error(ERROR_MODULE, [module, package, str(exc)]) from exc
        raise basic_error(ERROR_PACKAGE, [module, package, str(exc)]) from exc


CONNECT_ERROR = "Cannot connect to database({name!r}) with connection string {url}. You can test your connection with dbgen connect --test"


def test_connection(conn: 'Connection', name: str = ''):
    if not conn.test():
        bad_typer_print(CONNECT_ERROR.format(name='meta', url=conn.url()))
        raise typer.Exit(code=2)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Dict

from pydantic import BaseModel

if TYPE_CHECKING:
    from contextvars import Token  # pragma: no cover


class BaseContext(BaseModel):
    context_dict: Dict[str, Any]
    __context__: ContextVar
    _token: 'Token'

    def __enter__(self):
        token = self.__context__.set(self.context_dict)
        object.__setattr__(self, '_token', token)
        return self.context_dict

    def __exit__(self, *_):
        self.reset()

    def reset(self):
        self.__context__.reset(getattr(self, '_token'))

    @classmethod
    def get(cls):
        return cls.__context__.get(None)


class ModelContext(BaseContext):
    __context__: ContextVar = ContextVar('model')


class GeneratorContext(BaseContext):
    __context__: ContextVar = ContextVar('generator')



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import defaultdict
from datetime import date, datetime, time, timedelta
from decimal import Decimal
from enum import Enum
from typing import Dict, Optional, Set, Union
from uuid import UUID

import sqlalchemy.dialects.postgresql as postgres_types
import sqlalchemy.types as sa_types
import sqlmodel.sql.sqltypes as sqlmodel_types
from psycopg import adapters

from dbgen.core.base import Base
from dbgen.utils.typing import COLUMN_TYPE


class DataType(Base):
    """A container for mapping SQLAlchemy Types, python types, and postgres types."""

    name: str
    postgres_name: Optional[str] = None
    type_hint: type
    python_type: type
    column_type: COLUMN_TYPE
    _hashinclude_ = {'name'}

    class Config:
        """Pydantic doc string"""

        arbitrary_types_allowed = True

    def get_column_type(self) -> COLUMN_TYPE:
        return self.column_type

    def get_array_column_type(self) -> sa_types.ARRAY[sa_types.TypeEngine]:
        return sa_types.ARRAY(self.get_column_type())

    def get_python_type(self) -> type:
        return self.python_type

    @property
    def oid(self) -> int:
        key = self.postgres_name or self.name
        return adapters.types.get_oid(key)

    @property
    def array_oid(self) -> int:
        key = self.postgres_name or self.name
        return adapters.types.get_oid(f'{key}[]')

    @property
    def type_name(self):
        return self.postgres_name or self.name


class ColumnTypeRegistry:
    """A registry for all the datatypes in the model."""

    _registry: Dict[Union[str, int, COLUMN_TYPE], DataType]
    _default_python_type_registry: Dict[type, DataType]
    _python_type_registry: Dict[type, Set[DataType]]

    def __init__(self):
        self.clear()

    def clear(self):
        self._registry = {}
        self._default_python_type_registry = {}
        self._python_type_registry = defaultdict(set)

    def add(self, data_type: DataType) -> None:
        self._registry[data_type.name] = data_type
        self._registry[data_type.oid] = data_type
        self._registry[data_type.array_oid] = data_type
        self._registry[data_type.get_column_type()] = data_type
        self._registry[data_type.get_array_column_type()] = data_type
        # register default value as first value seen by registry
        if data_type.python_type not in self._python_type_registry:
            self._default_python_type_registry[data_type.python_type] = data_type
        self._python_type_registry[data_type.python_type].add(data_type)

    def __getitem__(self, key: Union[str, int, COLUMN_TYPE]) -> DataType:
        if isinstance(key, str):
            if key.endswith('[]'):
                key = key[:-2]
        try:
            if isinstance(key, sa_types.ARRAY):
                key = key.item_type
            if isinstance(key, sa_types.TypeEngine):
                key = type(key)
            return self._registry[key]
        except KeyError:
            raise KeyError(f"Cannot find key {key!r} in ColumnType Registry.")

    def get(self, key) -> Optional[DataType]:
        try:
            return self[key]
        except KeyError:
            return None

    def get_from_python_type(self, type_: type) -> DataType:
        if type_ in self._default_python_type_registry:
            return self._default_python_type_registry[type_]

        if getattr(type_, '__mro__', None):
            for type_curr in type_.__mro__:
                if type_curr in self._default_python_type_registry:
                    return self._default_python_type_registry[type_curr]

        if type_ not in self._default_python_type_registry:
            raise ValueError(f"Cannot find python type {type_!r} in registry")
        return self._default_python_type_registry[type_]

    def overwrite_default(self, data_type: DataType) -> None:
        """Overwrite the default DataType for a given python type."""
        self._default_python_type_registry[data_type.python_type] = data_type


# Order Matters here as the when two datatypes share the same python type the first one is selected
COLUMN_TYPES = [
    DataType(
        name='autostring',
        postgres_name='text',
        type_hint=str,
        python_type=str,
        column_type=sqlmodel_types.AutoString,
    ),
    DataType(name='text', type_hint=str, python_type=str, column_type=sa_types.Text),
    DataType(name='int4', type_hint=int, python_type=int, column_type=sa_types.Integer),
    DataType(name='int8', type_hint=int, python_type=int, column_type=sa_types.BigInteger),
    DataType(name='float8', type_hint=float, python_type=float, column_type=sa_types.Float),
    DataType(name='bool', type_hint=bool, python_type=bool, column_type=sa_types.Boolean),
    DataType(name='date', type_hint=date, python_type=date, column_type=sa_types.Date),
    DataType(
        name='timestamptz',
        type_hint=datetime,
        python_type=datetime,
        column_type=sa_types.DateTime(timezone=True),
    ),
    DataType(name='timestamp', type_hint=datetime, python_type=datetime, column_type=sa_types.DateTime),
    DataType(name='interval', type_hint=timedelta, python_type=timedelta, column_type=sa_types.Interval),
    DataType(name='time', type_hint=time, python_type=time, column_type=sa_types.Time),
    DataType(name='bytea', type_hint=bytes, python_type=bytes, column_type=sa_types.LargeBinary),
    DataType(
        name='guid', postgres_name='uuid', type_hint=UUID, python_type=UUID, column_type=sqlmodel_types.GUID
    ),
    DataType(name='uuid', type_hint=UUID, python_type=UUID, column_type=postgres_types.UUID),
    DataType(name='json', type_hint=dict, python_type=dict, column_type=sa_types.JSON),
    DataType(
        name='pg_json',
        postgres_name='json',
        type_hint=dict,
        python_type=dict,
        column_type=postgres_types.JSON,
    ),
    DataType(name='jsonb', type_hint=dict, python_type=dict, column_type=postgres_types.JSONB),
    DataType(name='numeric', type_hint=Decimal, python_type=Decimal, column_type=sa_types.Numeric),
    DataType(name='varchar', type_hint=str, python_type=str, column_type=sa_types.String),
    DataType(
        name='enum', postgres_name='varchar', type_hint=Enum, python_type=Enum, column_type=sa_types.Enum
    ),
]

# Initialize default registry and add the above columns to it
column_registry = ColumnTypeRegistry()

for data_type in COLUMN_TYPES:
    column_registry.add(data_type)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
from typing import TYPE_CHECKING
from typing import Generator as GenType
from typing import List, Optional, Tuple
from uuid import UUID

import typer
from sqlmodel import Session, func, select

from dbgen.configuration import config, get_engines, root_logger
from dbgen.core.metadata import GeneratorEntity, GeneratorRunEntity, RunEntity
from dbgen.core.run import RemoteGeneratorRun, RunConfig
from dbgen.utils.log import LogLevel, add_stdout_logger

if TYPE_CHECKING:
    from sqlalchemy.future import Engine  # pragma: no cover

generator_app = typer.Typer()


def get_runnable_gens(
    all_runs: bool = True, meta_engine: Optional['Engine'] = None
) -> GenType[Tuple[str, datetime, UUID], None, None]:
    statement = (
        select(  # type: ignore
            GeneratorEntity.name,
            func.max(GeneratorRunEntity.created_at),
            GeneratorEntity.id,
        )
        .join_from(GeneratorEntity, GeneratorRunEntity)
        .group_by(GeneratorEntity.name, GeneratorEntity.id)
        .order_by(GeneratorEntity.name, func.max(GeneratorRunEntity.created_at).desc())
    )
    if not all_runs:
        statement = statement.where(
            GeneratorRunEntity.run_id == select(func.max(RunEntity.id)).scalar_subquery()  # type: ignore
        )

    # If we don't have a metaengine grab one
    if not meta_engine:
        _, meta_engine = get_engines(config)

    with Session(meta_engine) as session:
        result = session.exec(statement.distinct())
        yield from result


@generator_app.command('list')
def list_generators(
    all_runs: bool = typer.Option(
        False, '-a', '--all', help='Print all possible generators not just most recent'
    )
):
    """List out the generators from the last run."""
    typer.echo("-" * 77)
    typer.echo(f"{'name':<20} {'last_run':<20} {'generator_id':20}")
    typer.echo("-" * 77)
    for gen_name, last_run, gen_id in get_runnable_gens(all_runs):
        typer.echo(f"{gen_name!r:<20} {last_run.strftime('%m/%d/%Y %H:%M:%S'):<20} {str(gen_id):<20}")


@generator_app.command('run')
def run_generator(
    gen_id: UUID,
    run_id: Optional[int] = typer.Argument(None),
    ordering: Optional[int] = typer.Argument(None),
    retry: bool = typer.Option(False, help="Ignore repeat checking"),
    level: LogLevel = typer.Option(LogLevel.INFO, help="Log level"),
    include: List[str] = typer.Option([], help="Generators to include"),
    exclude: List[str] = typer.Option([], help="Generators to xclude"),
    start: Optional[str] = typer.Option(None, help="Generator to start run at"),
    until: Optional[str] = typer.Option(None, help="Generator to finish run at."),
    batch: Optional[int] = typer.Option(None, help="Batch size for all generators"),
):
    # Retrieve the configured engines for
    main_engine, meta_engine = get_engines(config)
    run_config = RunConfig(
        retry=retry,
        start=start,
        until=until,
        exclude=exclude,
        include=include,
        progress_bar=False,
        batch_size=batch,
        log_level=level,
    )
    add_stdout_logger(root_logger, stdout_level=level)
    # Validate that the generator id provided is in the GeneratorEntity Table for remote running
    gens = {gen_id: gen_name for gen_name, _, gen_id in get_runnable_gens(meta_engine=meta_engine)}
    if gen_id not in gens:
        raise typer.BadParameter(f"Invalid generator id, no generator found with uuid {gen_id}.")

    # Need to validate the run_id to make sure that run has been initialized (i.e. a run exists with run_id)
    # and that a gen_run has not already been created with this gen_id and run_id
    if run_id is not None:
        with Session(meta_engine) as session:
            existing_gen_run = session.get(GeneratorRunEntity, (gen_id, run_id))
            if existing_gen_run:
                raise typer.BadParameter(
                    f"Found an existing generator_run with (generator_id, run_id) equal to ({gen_id},{run_id})"
                )
            run = session.get(RunEntity, run_id)
            if not run:
                raise typer.BadParameter(
                    f"No Run exists with run_id = {run_id} has the run been initialized?"
                )
    code = RemoteGeneratorRun(generator_id=gen_id).execute(
        main_engine, meta_engine, run_id, run_config, ordering
    )
    if code:
        raise typer.Exit(code=code)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Store useful python type hints for use in project."""
from typing import Any, Callable, Optional, Type, Union
from uuid import UUID

import sqlalchemy.types as sa_types

COLUMN_TYPE = Union[Type[sa_types.TypeEngine], sa_types.TypeEngine]
NoArgAnyCallable = Callable[[], Any]
IDType = Optional[UUID]



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""For parsing in config YAML."""
import os
import tempfile
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING, Optional, Tuple

from pydantic import BaseSettings, PostgresDsn, SecretStr, validator
from pydantic.tools import parse_obj_as

from dbgen.utils.log import setup_logger
from dbgen.utils.sql import Connection

if TYPE_CHECKING:
    from sqlalchemy.future import Engine  # pragma: no cover


# Force postgresql schemes for connection for sqlalchemy
class PostgresqlDsn(PostgresDsn):
    allowed_schemes = {"postgresql"}
    path: str


class ValidationEnum(str, Enum):
    COERCE = 'coerce'
    STRICT = 'strict'
    OFF = 'off'


class DBgenConfiguration(BaseSettings):
    """Settings for the pg4j, especially database connections."""

    main_dsn: PostgresqlDsn = parse_obj_as(PostgresqlDsn, "postgresql://postgres@localhost:5432/dbgen")
    main_schema: str = "public"
    main_password: SecretStr = ""  # type: ignore
    meta_dsn: Optional[PostgresqlDsn]
    meta_schema: str = "dbgen_log"
    meta_password: Optional[SecretStr] = None
    model_str: str = 'model.main:make_model'
    temp_dir: Path = Path(tempfile.gettempdir())
    validation: ValidationEnum = ValidationEnum.COERCE
    pdb: bool = False

    class Config:
        """Pydantic configuration"""

        env_file = os.environ.get("DBGEN_CONFIG", ".env")
        env_prefix = "DBGEN_"
        extra = "forbid"

    @validator('temp_dir')
    def validate_temp_dir(cls, temp_dir: Path) -> Path:
        if not temp_dir.exists():
            temp_dir.mkdir(exist_ok=True, parents=True)
        if temp_dir.is_dir():
            return temp_dir
        raise ValueError(f"temp_dir value {temp_dir} is not a directory")

    def display(self, show_defaults: bool = True, show_passwords: bool = False):
        params = [
            "dbgen_{} = {}".format(
                key,
                val.get_secret_value() if show_passwords and "password" in key else val,
            )
            for key, val in self.dict().items()
            if (show_defaults or key in self.__fields_set__) and val is not None
        ]
        params_str = "\n".join(params)
        output = f"""# DBgen Settings\n{params_str}"""
        return dedent(output)

    def __str__(self):
        return self.display()

    def __repr__(self):
        return self.display()


config = DBgenConfiguration()
root_logger = setup_logger()


def initialize(config_file: 'Path' = None) -> Tuple['Connection', 'Connection']:
    global config
    if config_file:
        config = update_config(config_file)
    return get_connections(config)


def update_config(config_file: 'Path') -> DBgenConfiguration:
    global config
    input_config = DBgenConfiguration(_env_file=config_file)
    new_config_kwargs = {**config.dict(), **input_config.dict()}
    config = DBgenConfiguration.parse_obj(new_config_kwargs)
    return config


def get_connections(config: DBgenConfiguration) -> Tuple['Connection', 'Connection']:
    main_conn = Connection.from_uri(config.main_dsn, config.main_schema, password=config.main_password)
    meta_dsn = config.meta_dsn or config.main_dsn
    meta_password = config.meta_password or config.main_password
    meta_conn = Connection.from_uri(meta_dsn, config.meta_schema, meta_password)
    return (main_conn, meta_conn)


def get_engines(config: DBgenConfiguration) -> Tuple['Engine', 'Engine']:
    (main_conn, meta_conn) = get_connections(config)
    return (main_conn.get_engine(), meta_conn.get_engine())



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
from functools import reduce
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from pydantic import Field, PrivateAttr
from pydantic.class_validators import validator
from sqlalchemy.future import Engine

from dbgen.core.args import Arg
from dbgen.core.base import Base
from dbgen.core.context import GeneratorContext, ModelContext
from dbgen.core.decorators import FunctionNode
from dbgen.core.dependency import Dependency
from dbgen.core.metadata import GeneratorEntity
from dbgen.core.node.extract import Extract
from dbgen.core.node.load import Load
from dbgen.core.node.query import BaseQuery
from dbgen.core.node.transforms import Transform
from dbgen.exceptions import DBgenMissingInfo
from dbgen.utils.graphs import topsort_with_dict

if TYPE_CHECKING:
    from networkx import DiGraph  # pragma: no cover

    from dbgen.core.node.computational_node import ComputationalNode  # pragma: no cover

list_field = Field(default_factory=lambda: [])

NAME_REGEX = re.compile(r'^[\w.-]+$')
DEFAULT_EXTRACT: Extract[dict] = Extract()


class Generator(Base):
    name: str
    description: str = "<no description>"
    extract: Union[BaseQuery, Extract] = Field(default_factory=Extract)
    transforms: List[Transform] = list_field
    loads: List[Load] = list_field
    tags: List[str] = list_field
    batch_size: Optional[int] = None
    additional_dependencies: Optional[Dependency] = None
    dependency: Optional[Dependency] = None
    _graph: Optional["DiGraph"] = PrivateAttr(None)
    _context: GeneratorContext = PrivateAttr(None)
    _hashexclude_ = {
        'dependency',
    }

    def __init__(self, name: str, **kwargs):
        super().__init__(name=name, **kwargs)
        gen_context = GeneratorContext.get()
        if not gen_context:
            self.validate_nodes()
        model_context = ModelContext.get()
        if model_context:
            model = model_context['model']
            model.add_gen(self)

    @validator('name')
    def validate_gen_name(cls, name):
        if not NAME_REGEX.match(name):
            raise ValueError(
                f"Generator names have to be alphanumeric characters, dashes, dots, and underscores. No spaces allowed!\n  Offending Name: {name}"
            )
        return name

    @validator('transforms', pre=True)
    def convert_functional_node(cls, transforms):
        return [val.pyblock if isinstance(val, FunctionNode) else val for val in transforms]

    def validate_nodes(self):
        nodes = self.loads + self.transforms + [self.extract]
        hashes = {node.hash for node in nodes}

        for node in nodes:
            for arg in node.inputs.values():
                if isinstance(arg, Arg) and arg.key not in hashes:
                    if arg.name.endswith('_id'):
                        hint = f"Arg name being asked for is '{arg.name}' which matches the pattern of a Load for Entity with name {arg.name[:-3]!r}. Are you missing a load?"
                    elif self.extract.hash == Extract().hash:
                        hint = "Generator is using the default extract, did you remember your query or extractor to the extract field?"
                    else:
                        hint = "The arg details seem to match a PyBlock, did you add all pyblocks?"
                    raise ValueError(
                        f"Generator(name={self.name!r}) encountered a validation error as a node is missing in the computational graph.\n  "
                        f"Node {node} is looking for an output named {arg.name!r} on another node with a hash {arg.key!r}\n  "
                        + hint
                    )

    def __str__(self) -> str:
        return f"Gen<{self.name}>"

    def __enter__(self) -> "Generator":
        self._context = GeneratorContext(context_dict={'generator': self})
        return self._context.__enter__()['generator']

    def __exit__(self, *args):
        self._context.__exit__(*args)
        self.validate_nodes()
        del self._context

    def _computational_graph(self) -> "DiGraph":
        if self._graph is None:
            from networkx import DiGraph

            nodes: Dict[str, "ComputationalNode"] = {self.extract.hash: self.extract}
            # Add transforms and loads
            nodes.update({transform.hash: transform for transform in self.transforms})
            nodes.update({load.hash: load for load in self.loads})
            # Add edges for every Arg in graph
            edges: List[Tuple[str, str]] = []
            for node_id, node in nodes.items():
                for key, arg in node.inputs.items():
                    if isinstance(arg, Arg):
                        if arg.key not in nodes:
                            raise DBgenMissingInfo(
                                f"Argument {key} of {node} refers to an object with a hash key {arg.key} asking for name \"{getattr(arg,'name','<No Name>')}\" that does not exist in the namespace.\n"
                                "Did you make sure to include all PyBlocks and Queries in the func kwarg of Generator()?"
                            )
                        edges.append((arg.key, node_id))

            graph = DiGraph()
            for node_id, node in nodes.items():
                graph.add_node(node_id, data=node)
            graph.add_edges_from(edges)
            self._graph = graph
        return self._graph

    def _sort_graph(self) -> List["ComputationalNode"]:
        graph = self._computational_graph()
        sorted_node_ids = topsort_with_dict(graph)
        sorted_nodes = [
            graph.nodes[key]["data"]
            for key in sorted_node_ids
            if not isinstance(graph.nodes[key]["data"], Extract)
        ]
        return [self.extract, *sorted_nodes]

    def _sorted_loads(self) -> List[Load]:
        sorted_nodes = self._sort_graph()
        return [node for node in sorted_nodes if isinstance(node, Load)]

    def _get_dependency(self) -> Dependency:
        dep_list = [self.additional_dependencies] if self.additional_dependencies else []
        dep_list.extend([node._get_dependency() for node in self._sort_graph()])
        self.dependency = reduce(lambda p, n: p.merge(n), dep_list, Dependency())
        return self.dependency

    def _get_gen_row(self) -> GeneratorEntity:
        # Assemble stringified dependency fields as we can't store sets in postgres easily
        deps = self._get_dependency()
        dep_kwargs = {}
        for x in deps.__fields__:
            dep_val = getattr(deps, x)
            dep_kwargs[x] = ",".join(dep_val) if dep_val else None
        return GeneratorEntity(
            id=self.uuid,
            name=self.name,
            description=self.description,
            tags=",".join(self.tags),
            query=self.extract.query if isinstance(self.extract, BaseQuery) else None,
            gen_json=self.serialize(),
            **dep_kwargs,
        )

    def run(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_id: int = None,
        ordering: int = None,
        run_config=None,
    ):
        from dbgen.core.run import GeneratorRun

        return GeneratorRun(generator=self).execute(
            main_engine,
            meta_engine,
            run_id,
            run_config,
            ordering,
        )

    def add_node(self, node: 'ComputationalNode') -> None:
        if isinstance(node, Extract):
            if self.extract.__class__ == DEFAULT_EXTRACT.__class__:
                self.extract = node
            else:
                raise ValueError(
                    f"Can only define 1 extractor per generator\n"
                    f"{self.extract} already defined defined on {self}\n"
                    f"Cannot add extract {node}"
                )
        elif isinstance(node, Transform):
            self.transforms.append(node)
        elif isinstance(node, FunctionNode):
            self.transforms.append(node.pyblock)
        elif isinstance(node, Load):
            self.loads.append(node)
        else:
            raise ValueError(f"Unknown Node Type {node} {type(node)}")



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from abc import ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Generic, TypeVar

from dbgen.core.base import Base
from dbgen.exceptions import DBgenMissingInfo

if TYPE_CHECKING:
    from dbgen.core.func import Func  # pragma: no cover
    from dbgen.core.node.transforms import PyBlock  # pragma: no cover

    Func


T = TypeVar('T')


class ArgLike(Base, Generic[T], metaclass=ABCMeta):
    @abstractmethod
    def arg_get(self, dic: dict):
        raise NotImplementedError

    def map(self, function: Callable[[Any], Any]) -> 'PyBlock':
        from dbgen.core.node.transforms import PyBlock

        return PyBlock(inputs=[self], function=function)


class Arg(ArgLike[T]):
    """
    How a function refers to a namespace
    """

    key: str
    name: str

    def __str__(self) -> str:
        return f"Arg({str(self.key)[:4]}...,{self.name})"

    def __iter__(self):
        raise TypeError(
            f"You are attempting to iterate/unpack the arg object {self}. This can commonly occur when a Extract or Transform outputs a single output when you expected two. Did you remember to set the outputs on your Extract or Transform?"
        )

    def arg_get(self, namespace: dict) -> T:
        """
        Common interface for Const and Arg to get values out of namespace
        """
        try:
            val = namespace[self.key][self.name]
            return val
        except KeyError:
            if self.key not in namespace:
                raise DBgenMissingInfo(
                    f"could not find hash, looking for output named '{self.name}' at this hash {self.key}"
                )
            else:
                err = "could not find '%s' in %s "
                raise DBgenMissingInfo(err % (self.name, list(namespace[self.key].keys())))


class Const(ArgLike[T]):
    val: T

    def __init__(self, val: T):
        super().__init__(val=val)

    def __str__(self) -> str:
        return f"Const<{self.val}>"

    def arg_get(self, _: dict) -> T:
        return self.val



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from pathlib import Path
from typing import Any, Dict, Generic, List, Mapping, Optional, Tuple, TypeVar, Union, overload

from pydantic import Field, validator

from dbgen.core.args import Arg, ArgLike, Const
from dbgen.core.base import Base
from dbgen.core.context import GeneratorContext
from dbgen.core.dependency import Dependency
from dbgen.exceptions import DBgenMissingInfo

BasicType = TypeVar('BasicType', int, float, str, bytes, Path)
T1 = TypeVar('T1')
T2 = TypeVar('T2')
T3 = TypeVar('T3')
T4 = TypeVar('T4')
T5 = TypeVar('T5')
Output = TypeVar('Output')


class ComputationalNode(Base, Generic[Output]):
    inputs: Mapping[str, Union[Const, Arg]] = Field(default_factory=lambda: {})
    outputs: List[str] = Field(default_factory=lambda: ["out"], min_items=1)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        gen_context = GeneratorContext.get()
        if gen_context:
            gen = gen_context['generator']
            gen.add_node(self)

    def _get_dependency(self) -> Dependency:
        return Dependency()

    @validator("inputs", pre=True)
    def convert_list_to_dict(cls, inputs):
        if isinstance(inputs, dict):
            return inputs
        new_inputs = {}
        for arg_idx, arg_val in enumerate(inputs):
            if isinstance(arg_val, (str, int, bool, float)):
                new_inputs[str(arg_idx)] = Const(val=arg_val)
            elif isinstance(arg_val, ArgLike):
                new_inputs[str(arg_idx)] = arg_val
            else:
                raise TypeError(f"Bad input type provided: {type(arg_val)} {arg_val}")
        return new_inputs

    @validator("outputs", pre=True)
    def unique_keys(cls, outputs):
        if isinstance(outputs, set):
            raise ValueError(
                f"Outputs cannot be a set as order cannot be preserved...{outputs}",
            )
        elif isinstance(outputs, (list, tuple)):
            if len(set(outputs)) != len(outputs):
                raise ValueError(["No duplicate output names allowed"])
        return outputs

    @property
    def name(self):
        return str(self)

    def __getitem__(self, key: str):
        assert key in self.outputs, f"Key {key} not found: {self.outputs}"
        return Arg(key=self.hash, name=key)

    def _get_inputs(self, namespace: Dict[str, Mapping[str, Any]]) -> Dict[str, Any]:
        try:
            input_variables = {name: value.arg_get(namespace) for name, value in self.inputs.items()}
        except (TypeError, IndexError) as e:
            print(e)
            print(namespace)
            raise ValueError()
        except AttributeError:
            invalid_args = [getattr(arg, "arg_get", None) is None for arg in self.inputs]
            missing_args = filter(lambda x: invalid_args[x], range(len(invalid_args)))
            raise DBgenMissingInfo(
                f"Argument(s) {' ,'.join(map(str,missing_args))} to {self.name} don't have arg_get attribute:\n Did you forget to wrap a Const around a PyBlock Argument?"
            )
        return input_variables

    def run(self, namespace: Dict[str, Mapping[str, Any]]) -> Optional[Dict[str, Any]]:
        return {}

    @overload
    def results(
        self: 'ComputationalNode[Tuple[T1,T2,T3,T4,T5]]',
    ) -> Tuple[Arg[T1], Arg[T2], Arg[T3], Arg[T4], Arg[T5]]:
        ...

    @overload
    def results(self: 'ComputationalNode[Tuple[T1,T2,T3,T4]]') -> Tuple[Arg[T1], Arg[T2], Arg[T3], Arg[T4]]:
        ...

    @overload
    def results(self: 'ComputationalNode[Tuple[T1,T2,T3]]') -> Tuple[Arg[T1], Arg[T2], Arg[T3]]:
        ...

    @overload
    def results(self: 'ComputationalNode[Tuple[T1,T2]]') -> Tuple[Arg[T1], Arg[T2]]:
        ...

    @overload
    def results(self: 'ComputationalNode[Tuple[T1]]') -> Arg[T1]:
        ...

    @overload
    def results(self: 'ComputationalNode[BasicType]') -> Arg[BasicType]:
        ...

    def results(self):
        arglist = tuple(iter(map(self.__getitem__, self.outputs)))
        return arglist[0] if len(arglist) == 1 else arglist



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import logging
from datetime import datetime
from functools import partial
from io import StringIO
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)
from uuid import UUID

from sqlalchemy import Column, DateTime
from sqlalchemy.orm import registry
from sqlalchemy.sql import func
from sqlalchemy.sql.base import ImmutableColumnCollection
from sqlalchemy.sql.schema import Table
from sqlmodel.main import Field, FieldInfo, SQLModel, SQLModelMetaclass

from dbgen.configuration import config as dbgen_config
from dbgen.core.args import ArgLike, Const
from dbgen.core.attribute import Attribute
from dbgen.core.base import Base, BaseMeta
from dbgen.core.node.load import Load, LoadEntity
from dbgen.core.type_registry import column_registry
from dbgen.exceptions import DBgenMissingInfo, InvalidArgument


def inherit_field(
    bases, field_name: str, initial_value=set(), joiner=lambda x, y: x.union(y), type_check: bool = True
):
    field_val = initial_value
    for base in reversed(bases):
        curr_id = getattr(base, field_name, initial_value)
        if curr_id is not None:
            if type_check and not isinstance(curr_id, type(initial_value)):
                raise TypeError(f"Invalid {field_name} val: {curr_id}")
            field_val = joiner(field_val, curr_id)
    return field_val


overwrite_parent = partial(inherit_field, initial_value="", joiner=lambda x, y: y)
DEFAULT_ENTITY_REGISTRY = registry()
logger = logging.getLogger('dbgen.core.entity')

_T = TypeVar("_T")


def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]:
    return lambda a: a


valid_kwargs_names = {
    '__identifying__',
    '__annotations__',
    '_hashexclude_',
    '_hashinclude_',
    'table',
    'registry',
    'all_identifying',
    'force_validation',
}


@__dataclass_transform__(
    kw_only_default=True,
    field_descriptors=(
        Field,
        FieldInfo,
        Attribute,
    ),
)
class EntityMetaclass(SQLModelMetaclass, BaseMeta):
    def __new__(mcs, name, bases, attrs, **kwargs):
        # validate kwargs passed in to catch user input errors
        invalid_kwargs = [name for name in kwargs if name not in valid_kwargs_names]
        if invalid_kwargs:
            raise InvalidArgument(
                f"Unknown kwargs {invalid_kwargs} passed to Entity {name}. The only valid kwargs are {valid_kwargs_names}"
            )
        # Join the keys from all parents for __identifying__, _hashinclude_, and _hashexclude_
        new_attrs = attrs.copy()
        for value in ("__identifying__", "_hashexclude_", "_hashinclude_"):
            starting = new_attrs.get(value, set())
            if isinstance(starting, list):
                starting = set(starting)
            new_attrs[value] = starting.union(inherit_field(bases, value))

        if kwargs.get('all_identifying', False):
            assert (
                "__identifying__" not in attrs
            ), f"Error with Entity {name}. Can't supply both all_identifying kwarg and __identifying__ attr"
            new_attrs['__identifying__'] = new_attrs['__identifying__'].union(
                {key for key in attrs.get('__annotations__', {})}
            )
        # Automatically add identifying attributes to the hashinclude
        new_attrs["_hashinclude_"].update(new_attrs.get("__identifying__"))
        # Set the default registry to be the default_registry
        if "registry" not in kwargs:
            kwargs["registry"] = DEFAULT_ENTITY_REGISTRY

        # Call SQLModelMetaclass.__new__
        cls = super().__new__(mcs, name, bases, new_attrs, **kwargs)
        # Validate that we don't have table=True on current class and a base
        current_cls_is_table = getattr(cls.__config__, "table", False) and kwargs.get("table")
        setattr(cls, "_is_table", current_cls_is_table)

        if current_cls_is_table:
            base_is_table = False
            for base in bases:
                config = getattr(base, "__config__", None)
                if config and getattr(config, "table", False):
                    base_is_table = True
                    offending_base_name = base.__name__
                    break
            if base_is_table:
                raise ValueError(
                    "Can't use table=True when inheriting from another table.\n"
                    f"Both {offending_base_name} and {name} have table=True set.\n"
                    "Create a common ancestor with table=False and mutaually inherit from that."
                )
        # Need to look into parents to find schema, only using most recent
        schema_key = "__schema__"
        schema = getattr(cls, schema_key, "") or overwrite_parent(bases, schema_key)
        table_args = getattr(cls, "__table_args__", None) or dict().copy()
        if not schema:
            schema = dbgen_config.main_schema
        if schema:
            setattr(cls, schema_key, schema)
            table_args = table_args.copy()
            table_args.update({"schema": schema})

        setattr(cls, "__table_args__", table_args)
        setattr(
            cls,
            "__fulltablename__",
            f"{schema}.{cls.__tablename__}" if schema else cls.__tablename__,
        )
        # Validate __identifying__ by making sure all attribute exists on Entity
        unknown_ids = list(
            filter(
                lambda x: x not in cls.__fields__,
                new_attrs["__identifying__"],
            )
        )
        if unknown_ids:
            raise ValueError(
                f"Invalid Entity Class Definition. Identifying attributes not found on class: {unknown_ids}"
            )
        return cls

    def __init__(cls, name, bases, attrs, **kwargs):
        if cls._is_table:
            registry = cls._sa_registry
            if cls.__fulltablename__ in registry.metadata.tables:
                raise ValueError(
                    f"The Class {attrs.get('__module__','')}.{name}'s __table_name__ {cls.__tablename__!r} already present in the registry's metadata.\n"
                    "This can occur if two Entity sub-classes share a case-insensitive name or if the same table has been added to the registry twice.\n"
                    "To address this you can set a different __tablename__ attribute for one or to clear the registry, you can call Entity.clear_registry() prior to declaring this class."
                )
        super().__init__(name, bases, attrs, **kwargs)


class BaseEntity(Base, SQLModel, metaclass=EntityMetaclass):
    __identifying__: ClassVar[Set[str]]
    __fulltablename__: ClassVar[str]
    __schema__: ClassVar[str]
    __table__: ClassVar[Table]
    _is_table: ClassVar[bool]
    _sa_registry: ClassVar[registry]

    class Config:
        """Pydantic Config"""

        force_validation = True

    @classmethod
    def _columns(cls) -> ImmutableColumnCollection:
        if isinstance(cls.__fulltablename__, str):
            table = cls.metadata.tables.get(cls.__fulltablename__)
            if table is not None:
                return table.c
            raise ValueError(
                f"{cls.__fulltablename__} not in metadata, is table=True set? {cls.metadata.tables}"
            )
        raise ValueError(f"Can't read __fulltablename__ {cls.__fulltablename__}")

    @classmethod
    def _get_load_entity(cls) -> LoadEntity:
        """Returns a LoadEntity which has the bare-minimum needed to load into this table."""
        # Check that entity is a table
        if not cls._is_table:
            raise ValueError(f"{cls.__qualname__} is not a table. Can't get LoadEntity of a non-table Entity")
        columns = cls._columns()
        # Search for primary key name
        primary_keys = [x.name for x in cls.__table__.primary_key]
        if len(primary_keys) > 1:
            raise NotImplementedError(f"Multiple primary_keys found: {primary_keys}")
        elif not primary_keys:
            raise ValueError(f"No primary key found on {cls.__name__}'s columns:\n{columns}")
        primary_key_name = primary_keys[0]
        all_attrs = {col.name: col for col in columns if not col.foreign_keys}
        all_fks = {col.name: col for col in columns if col.foreign_keys}

        # Create the attribute dict which maps attribute name to column type
        attributes: Dict[str, str] = {}
        required: Set[str] = set()
        for col_name, col in columns.items():
            try:
                dt = column_registry[col.type]
                field = cls.__fields__[col_name]
                # append [] to the type name if the column is an array type
                attributes[col_name] = (
                    f"{dt.type_name}[]" if getattr(col.type, '_is_array', False) else dt.type_name
                )
                # Check for required fields
                if field.required:
                    required.add(col_name)

            except KeyError:
                raise TypeError(
                    f"Cannot parse column {col_name} on table {cls.__tablename__} due to its unknown type {type(col.type)}"
                )
        foreign_keys = set(all_fks.keys())
        identifying_attributes = {x for x in all_attrs if x in cls.__identifying__}
        identifying_fks = [x for x in all_fks if x in cls.__identifying__]
        return LoadEntity(
            name=cls.__tablename__ or cls.__name__,
            schema_=cls.__schema__,
            entity_class_str=f"{cls.__module__}.{cls.__qualname__}",
            primary_key_name=primary_key_name,
            attributes=attributes,
            required=required,
            foreign_keys=foreign_keys,
            identifying_attributes=identifying_attributes,
            identifying_foreign_keys=identifying_fks,
        )

    @classmethod
    def load(cls, insert: bool = False, validation: Optional[str] = None, **kwargs) -> Load[UUID]:
        name = cls.__tablename__
        assert isinstance(name, str)
        # TODO check if we need this anymore
        key_filter = lambda keyval: keyval[0] != "insert" and not isinstance(keyval[1], (ArgLike, Load))
        invalid_args = list(filter(key_filter, kwargs.items()))
        JSONAble = (str, int, float, dict, tuple)
        for arg_name, invalid_arg in invalid_args:
            # Check Invalid args to see if a const block would be appropriate
            if isinstance(invalid_arg, JSONAble):
                kwargs[arg_name] = Const(invalid_arg)
            else:
                raise ValueError(f"Non-jsonable constant value found: {arg_name}\n{invalid_arg}")

        # get PK
        pk = kwargs.pop('id', None)
        # if we don't have a PK reference check for missing ID info
        if not pk:
            missing = cls.__identifying__ - set(kwargs)
            if missing:
                err = (
                    f"Cannot refer to a row in {name} without a PK or essential data."
                    f" Missing essential data: {missing}\n"
                    f" Did you provide the primary_key to the wrong column name? the correct column name is the table name (i.e. {name}=...)"
                )
                raise DBgenMissingInfo(err)
        if insert:
            required_fields = {key for key, val in cls.__fields__.items() if val.required}
            missing_required = required_fields - set(kwargs)
            if missing_required:
                err = (
                    "Cannot insert a row into table {!r} without a required data."
                    " Missing required data: {}"
                )
                raise DBgenMissingInfo(err.format(name, missing_required))
        # Iterate through the columns to ensure we have no unknown kwargs
        class_columns: List[Column] = list(cls._columns()) or []
        all_attrs = {col.name: col for col in class_columns if not col.foreign_keys}
        all_fks = {col.name: col for col in class_columns if col.foreign_keys}
        attrs = {key: val for key, val in kwargs.items() if key in all_attrs}
        fks = {key: col for key, col in kwargs.items() if key not in attrs}
        for fk in fks:
            if fk not in all_fks:
                raise InvalidArgument(f'unknown "{fk}" kwarg in Load of {name}')

        for k, v in fks.items():
            if isinstance(v, Load):
                fks[k] = v[v.outputs[0]]

        return Load(
            load_entity=cls._get_load_entity(),
            primary_key=pk,
            inputs={**attrs, **fks},
            insert=insert,
            validation=validation,
        )

    @classmethod
    def _quick_load(cls, connection, rows: Iterable[Iterable[Any]], column_names: List[str]) -> None:
        """Bulk load many rows into entity"""
        from dbgen.templates import jinja_env

        # Assemble rows into stringio for copy_from statement
        io_obj = StringIO()
        for row in rows:
            io_obj.write("\t".join(map(str, row)) + "\n")
        io_obj.seek(0)

        # Temporary table to copy data into
        # Set name to be hash of input rows to ensure uniqueness for parallelization
        temp_table_name = f"{cls.__tablename__}_temp_load_table"

        load_entity = cls._get_load_entity()
        # Need to create a temp table to copy data into
        # Add an auto_inc column so that data can be ordered by its insert location
        drop_temp_table = f"DROP TABLE IF EXISTS {temp_table_name};"
        create_temp_table = """
        CREATE TEMPORARY TABLE {temp_table_name} AS
        TABLE {schema}.{obj}
        WITH NO DATA;
        ALTER TABLE {temp_table_name}
        ADD COLUMN auto_inc SERIAL NOT NULL;
        """.format(
            obj=load_entity.name,
            schema=load_entity.schema_,
            temp_table_name=temp_table_name,
        )

        insert_template = jinja_env.get_template("insert.sql.jinja")
        template_args = dict(
            obj=load_entity.name,
            obj_pk_name=load_entity.primary_key_name,
            temp_table_name=temp_table_name,
            all_column_names=column_names,
            schema=load_entity.schema_,
            first=False,
            update=True,
        )
        insert_statement = insert_template.render(**template_args)
        with connection.cursor() as curs:
            curs.execute(drop_temp_table)
        connection.commit()
        with connection.cursor() as curs:
            curs.execute(create_temp_table)
            curs.copy_from(io_obj, temp_table_name, null="None", columns=column_names)
            curs.execute(insert_statement)
        connection.commit()
        with connection.cursor() as curs:
            curs.execute(drop_temp_table)
        connection.commit()

    @classmethod
    def clear_registry(cls):
        """Removes all Entity classes from the Entity registry"""
        cls.metadata.clear()
        cls._sa_registry.dispose()

    @classmethod
    def foreign_key(cls, primary_key: bool = False):
        """Removes all Entity classes from the Entity registry"""
        load_entity = cls._get_load_entity()
        return Field(
            None,
            foreign_key=f"{cls.__fulltablename__}.{load_entity.primary_key_name}",
            primary_key=primary_key,
            sa_column_kwargs={'index': True},
        )


id_field = Field(
    default=None,
    primary_key=True,
    sa_column_kwargs={"autoincrement": False, "unique": True},
)
gen_id_field = Field(
    default=None,
)

get_created_at_field = lambda: Field(
    None, sa_column=Column(DateTime(timezone=True), server_default=func.now())
)


class Entity(BaseEntity):
    id: Optional[UUID] = id_field
    gen_id: Optional[UUID]
    created_at: Optional[datetime] = get_created_at_field()


Model = TypeVar("Model", bound="BaseEntity")


@overload
def create_entity(
    model_name: str,
    field_definitions: Dict[str, Union[Tuple[type, Any], type, Tuple[type, ...]]],
    base: None = None,
    identifying: Set[str] = None,
    schema: Optional[str] = None,
    __module__: str = __name__,
    **kwargs,
) -> Type[BaseEntity]:
    ...


@overload
def create_entity(
    model_name: str,
    field_definitions: Dict[str, Union[Tuple[type, Any], type, Tuple[type, ...]]],
    base: Type[Model],
    identifying: Set[str] = None,
    schema: Optional[str] = None,
    __module__: str = __name__,
    **kwargs,
) -> Type[Model]:
    ...


def create_entity(
    model_name: str,
    field_definitions: Dict[str, Union[Tuple[type, Any], type, Tuple[type, ...]]] = None,
    base: Optional[Type[Model]] = None,
    identifying: Set[str] = None,
    schema: Optional[str] = None,
    __module__: str = __name__,
    **kwargs,
) -> Type[Model]:
    """
    Dynamically create a model, similar to the Pydantic `create_model()` method
    :param model_name: name of the created model
    :param field_definitions: data fields of the create model
    :param base: base to inherit from
    :param __module__: module of the created model
    :param **kwargs: Other keyword arguments to pass to the metaclass constructor, e.g. table=True
    """

    if base is None:
        base = cast(Type["Model"], BaseEntity)
    field_definitions = field_definitions or {}

    fields = {}
    annotations = {}
    identifying = identifying or set()

    for f_name, f_def in field_definitions.items():
        if f_name.startswith("_"):
            raise ValueError("Field names may not start with an underscore")
        try:
            if isinstance(f_def, tuple) and len(f_def) > 1:
                f_annotation, f_value = f_def
            elif isinstance(f_def, tuple):
                f_annotation, f_value = f_def[0], Field(nullable=False)
            else:
                f_annotation, f_value = f_def, Field(nullable=False)
        except ValueError as e:
            raise ValueError(
                "field_definitions values must be either a tuple of (<type_annotation>, <default_value>)"
                "or just a type annotation [or a 1-tuple of (<type_annotation>,)]"
            ) from e

        if f_annotation:
            annotations[f_name] = f_annotation
        fields[f_name] = f_value

    namespace = {
        "__annotations__": annotations,
        "__identifying__": identifying,
        "__module__": __module__,
    }
    if schema is not None:
        namespace.update({"__schema__": schema})
    if "registry" in kwargs:
        assert isinstance(kwargs.get("registry"), registry), "Invalid type for registry:"
    namespace.update(fields)  # type: ignore

    return EntityMetaclass(model_name, (base,), namespace, **kwargs)  # type: ignore



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Welcome to DBgen!"""

__author__ = "Michael Statt"
__email__ = "michael.statt@modelyst.io"
__maintainer__ = "Michael Statt"
__maintainer_email__ = "michael.statt@modelyst.io"
__version__ = "0.5.16"


from sqlalchemy.orm import registry

# External packages that we reexport for convenience
from sqlmodel import select

from dbgen.core.args import Const
from dbgen.core.decorators import transform
from dbgen.core.entity import BaseEntity, Entity
from dbgen.core.func import Env, Import
from dbgen.core.generator import Generator
from dbgen.core.model import Model
from dbgen.core.node.extract import Extract
from dbgen.core.node.query import Query
from dbgen.core.node.transforms import PyBlock
from dbgen.utils.typing import IDType



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import contextlib
import subprocess
from enum import Enum
from io import StringIO
from pathlib import Path
from typing import Optional

import typer

import dbgen.cli.styles as styles
from dbgen.cli.generator import generator_app
from dbgen.cli.model import model_app
from dbgen.cli.new import new_app
from dbgen.cli.options import config_option
from dbgen.cli.run import run_app
from dbgen.configuration import get_connections, update_config
from dbgen.utils.misc import which

app = typer.Typer(no_args_is_help=True)
app.add_typer(generator_app, name='generator')
app.add_typer(run_app, name='run')
app.add_typer(model_app, name='model')
app.add_typer(new_app, name='new')


@app.command()
def print_logo():
    typer.echo(styles.LOGO_STYLE)


app.command("version")(lambda: typer.echo(styles.LOGO_STYLE))


@app.command(name="config")
def get_config(
    config_file: Optional[Path] = config_option,
    show_password: bool = False,
    show_defaults: bool = True,
    out_pth: Optional[Path] = typer.Option(
        None,
        '--out',
        '-o',
        help="Location to write parametrized config",
    ),
):
    """
    Prints out the configuration of dbgen given an optional config_file or using the envvar DBGEN_CONFIG
    """

    typer.echo(styles.LOGO_STYLE)
    # Initialize settings to get the connections

    # Print config to stdout
    styles.delimiter()
    # Notify of config file
    if config_file:
        typer.echo(f"Config File found at location: {config_file.absolute()}")
        config = update_config(config_file)
    else:
        typer.echo("No config file found using defaults.")

    # If out_pth provided write the current config to the path provided and return
    if out_pth:
        with open(out_pth, "w") as f:
            f.write(config.display(True, True))

    styles.delimiter()
    typer.echo(config.display(show_defaults, show_password))


class DBgenDatabase(str, Enum):
    """enum for setting database to connect to"""

    META = "meta"
    MAIN = "main"


@app.command(name="connect")
def test_conn(
    connect: DBgenDatabase = typer.Argument(DBgenDatabase.MAIN, help="Database to connect the meta or main."),
    config_file: Optional[Path] = config_option,
    test: bool = typer.Option(False, "-t", "--test", help="Test the main and metadb connections"),
    show_password: bool = typer.Option(
        False, "-p", "--show-password", help="Expose password in printed dsn when testing."
    ),
):
    """
    Prints out the configuration of dbgen given an optional config_file or using the envvar DBGEN_CONFIG
    """

    if config_file:
        config = update_config(config_file)
    main_conn, meta_conn = get_connections(config)

    # If connect is chosen connect to the database selected using CLI sql
    if connect and not test:
        # set the engine based on connect string provided
        conn = main_conn if connect == DBgenDatabase.MAIN else meta_conn
        # Test connection to database
        if not conn.test():
            styles.bad_typer_print(f"Cannot connect to {str(connect)} db")
            raise typer.Exit(2)
        # Attempt to use psql and pgcli to connect to database
        try:
            # Filter out executibles using which function
            exes = list(filter(lambda x: x is not None, map(which, ("pgcli", "psql"))))
            # If we find no executables exit
            if not exes:
                styles.bad_typer_print(
                    "Cannot find either psql or pgcli in $PATH. Please install them to connect to database."
                )
                raise typer.Exit(2)
            # If we have valid executible run the command with the dsn provided
            subprocess.check_call(
                [exes[0], conn.url(False, True)],
            )
        except subprocess.CalledProcessError as exc:
            styles.bad_typer_print("Error connecting!")
            styles.bad_typer_print(exc)
        # Quit once finished
        raise typer.Exit()

    styles.delimiter()
    for conn, label in zip((main_conn, meta_conn), ("Main", "Meta")):
        styles.good_typer_print(f"Checking {label} DB...")
        new_stdout = StringIO()
        with contextlib.redirect_stdout(new_stdout):
            check = conn.test()
        test_output = "\n".join(new_stdout.getvalue().strip().split("\n")[1:])
        if check:
            styles.good_typer_print(
                f"Connection to {label} DB at {conn.url(not show_password,True)} all good!"
            )
            if test_output:
                styles.good_typer_print(test_output)
        else:
            styles.bad_typer_print(f"Cannot connect to {label} DB at {conn.url(not show_password,True)}!")
            if test_output:
                styles.bad_typer_print("Error Message:")
                styles.bad_typer_print("\n".join(["\t" + line for line in test_output.split("\n")]))
        styles.delimiter()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['dbgen',
 'dbgen.cli',
 'dbgen.core',
 'dbgen.core.node',
 'dbgen.providers',
 'dbgen.providers.aws',
 'dbgen.templates',
 'dbgen.utils']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=3.0.1,<4.0.0',
 'SQLAlchemy>=1.4.25,<2.0.0',
 'boto3>=1.18.42,<2.0.0',
 'modelyst-sqlmodel>=0.0.7,<0.0.8',
 'networkx>=2.6.3,<3.0.0',
 'psutil>=5.9.0,<6.0.0',
 'psycopg2-binary>=2.9.1,<3.0.0',
 'psycopg>=3.0.1,<4.0.0',
 'pydantic[dotenv]>=1.8.2,<2.0.0',
 'pydasher>=0.0.13,<0.0.14',
 'tqdm>=4.62.2,<5.0.0',
 'typer>=0.4.0,<0.5.0',
 'typing-extensions>=3.10.0.1']

extras_require = \
{'cookiecutter': ['cookiecutter>=1.7.3,<2.0.0']}

entry_points = \
{'console_scripts': ['dbgen = dbgen.__main__:main']}

setup_kwargs = {
    'name': 'modelyst-dbgen',
    'version': '0.5.16',
    'description': 'DBgen (Database Generator) is an open-source Python library for connecting raw data, scientific theories, and relational databases',
    'long_description': '<!--\n   Copyright 2021 Modelyst LLC\n\n   Licensed under the Apache License, Version 2.0 (the "License");\n   you may not use this file except in compliance with the License.\n   You may obtain a copy of the License at\n\n       http://www.apache.org/licenses/LICENSE-2.0\n\n   Unless required by applicable law or agreed to in writing, software\n   distributed under the License is distributed on an "AS IS" BASIS,\n   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.\n   See the License for the specific language governing permissions and\n   limitations under the License.\n -->\n\n# DBgen\n\n<p align="center">\n  <a href="https://dbgen.modelyst.com"><img src="docs/img/dbgen_logo.png" alt="DBgen"></a>\n</p>\n\n<p align="center">\n   <a href="https://github.com/modelyst/dbgen   /actions?query=workflow%3ATest" target="_blank">\n      <img src="https://github.com/modelyst/dbgen/workflows/Test/badge.svg" alt="Test">\n   </a>\n   <a href="https://github.com/modelyst/dbgen/actions?query=workflow%3APublish" target="_blank">\n      <img src="https://github.com/modelyst/dbgen/workflows/Publish/badge.svg" alt="Publish">\n   </a>\n   <a href="https://github.com/modelyst/dbgen/actions/workflows/publish_docs.yml" target="_blank">\n      <img src="https://github.com/modelyst/dbgen/actions/workflows/publish_docs.yml/badge.svg">\n   </a>\n   <a href="https://codecov.io/gh/modelyst/dbgen">\n      <img src="https://codecov.io/gh/modelyst/dbgen/branch/master/graph/badge.svg?token=V4I8PPUIBU"/>\n   </a>\n   <a href="https://codecov.io/gh/modelyst/dbgen">\n      <img src="docs/img/interrogate.svg"/>\n   </a>\n</p>\n---\n\n**Documentation**: <a href="https://dbgen.modelyst.com" target="_blank">https://dbgen.modelyst.com</a>\n\n**Github**: <a href="https://github.com/modelyst/dbgen" target="_blank">https://github.com/modelyst/dbgen</a>\n\n---\n\n:exclamation:  Please note that this project is actively under major rewrites and installations are subject to breaking changes.\n\n---\nDBgen (Database Generator) is an open-source Python library for\nconnecting raw data, scientific theories, and relational databases.\nThe package was designed with a focus on the developer experience at the core.\nDBgen was initially developed by [Modelyst](https://www.modelyst.com/).\n\n## What is DBgen?\n\nDBgen was designed to support scientific data analysis with the following\ncharacteristics:\n\n1.  Transparent\n\n    - Because scientific efforts ought be shareable and mutually\n      understandable.\n\n2.  Flexible\n\n    - Because scientific theories are under continuous flux.\n\n3.  Maintainable\n    - Because the underlying scientific models one works with are\n      complicated enough on their own, we can\'t afford to introduce\n      any more complexity via our framework.\n\nDBGen is an opinionated ETL tool. While many other ETL tools exist, they rarely\ngive the tools necessary for a scientific workflow.\nDBGen is a tool that helps populate a single postgresql database using a transparent, flexible, and mainatable data pipeline.\n\n### Alternative tools\n\nOrchestrators: Many tools exist to orchestrate python workflows. However, these tools often often are too general to help the average scientist wrangle their data or are so specific to storing a given workflow type they lack the flexibility needed to address the specifics of a scientist\'s data problems. Many other tools also come packaged with powerful\n#### General Orchestration Tools\n1. [Airflow](https://airflow.apache.org/)\n2. [Prefect](https://www.prefect.io/)\n3. [Luigi](https://github.com/spotify/luigi)\n\n#### Computational Science Workflow Tools\n1. [Fireworks](https://materialsproject.github.io/fireworks/)\n2. [AiiDA](http://www.aiida.net/)\n3. [Atomate](https://atomate.org/)\n\n## What isn\'t DBgen?\n\n1. An [ORM](https://en.wikipedia.org/wiki/Object-relational_mapping) tool (see [Hibernate](http://hibernate.org/orm/) for Java or [SQLAlchemy](https://www.sqlalchemy.org/) for Python)\n\n   - DBGen utilizes the popular SQLAlchemy ORM to operate at an even higher level extraction, allowing the users to build pipelines and schema without actively thinking about the database tables or insert and select statements required to connect the workflow together.\n\n2. A database manager (see\n   [MySQLWorkbench](https://www.mysql.com/products/workbench/),\n   [DBeaver](https://dbeaver.io/), [TablePlus](https://tableplus.com/),\n   etc.)\n3. An opinionated tool with a particular schema for scientific data /\n   theories.\n\n## Getting DBgen\n\n### Via Github\n\nCurrently, the only method of installing DBgen is through Github. This is best done by using the [poetry](https://python-poetry.org/) package manager. To do this, first clone the repo to a local directory. Then use the command `poetry install` in the directory to install the required dependencies. You will need at least python 3.7 to install the package.\n```Bash\n# Get DBgen\ngit clone https://github.com/modelyst/dbgen\ncd ./dbgen\n# Get Poetry\ncurl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python3 -\n# Install Poetrywhich ma\npoetry install\npoetry shell\n# Test dbgen\ndbgen serialize dbgen.example.main:make_model\n```\n### Via Pip\n```Bash\npip install modelyst-dbgen\n```\n\n### API documentation\n\nDocumentation of modules and classes can be found in\nAPI docs \\</modules\\>.\n\n#### Reporting bugs\n\nPlease report any bugs and issues at DBgen\'s [Github Issues\npage](https://github.com/modelyst/dbgen/issues).\n\n## License\n\nDBgen is released under the [Apache 2.0 License](license/).\n',
    'author': 'Michael Statt',
    'author_email': 'michael.statt@modelyst.io',
    'maintainer': 'Michael Statt',
    'maintainer_email': 'michael.statt@modelyst.io',
    'url': 'https://www.dbgen.modelyst.com',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from pprint import pformat
from typing import Any, Callable, Dict, List, Set, Tuple, Union
from uuid import UUID

from networkx import DiGraph, NetworkXUnfeasible
from networkx.algorithms import lexicographical_topological_sort, simple_cycles
from pydantic import BaseModel


# Graph
# --------
def topsort_with_dict(G: DiGraph) -> List:
    """
    Assuming a graph with object names and dict mapping names to objects,
    perform a topsort and return the list of objects.
    """
    try:
        sortd = list(lexicographical_topological_sort(G))
        return sortd
    except NetworkXUnfeasible:
        cycles = pformat(list(simple_cycles(G)))
        raise ValueError(f"Cycles found: {cycles}")


class SerializedNode(BaseModel):
    id: UUID
    name: str
    dependency: Dict[str, Set[str]]


class SerializedGraph(BaseModel):
    metadata: Dict[str, Union[str, float, int, None]]
    nodes: Dict[str, SerializedNode]
    edges: List[Tuple[str, str]]


def serialize_graph(
    graph,
    node_serializer: Callable[[Any], SerializedNode],
    metadata: Dict[str, Union[str, float, int, None]] = None,
) -> Dict[str, Any]:
    metadata = metadata or {}
    output: SerializedGraph = SerializedGraph(nodes={}, edges=[], metadata=metadata)
    for key, data in graph.nodes.items():
        node = data
        output.nodes[key] = node_serializer(node)
    for source, target in graph.edges:
        assert isinstance(source, str) and isinstance(
            target, str
        ), f"Edge keys not strings! {type(source)} {type(target)}"
        output.edges.append((source, target))
    return output.dict()


# def topological_sort(G, key=None):
#     """Returns a generator of nodes in lexicographically topologically sorted
#     order.

#     A topological sort is a nonunique permutation of the nodes such that an
#     edge from u to v implies that u appears before v in the topological sort
#     order.

#     Parameters
#     ----------
#     G : NetworkX digraph
#         A directed acyclic graph (DAG)

#     key : function, optional
#         This function maps nodes to keys with which to resolve ambiguities in
#         the sort order.  Defaults to the identity function.

#     Returns
#     -------
#     iterable
#         An iterable of node names in lexicographical topological sort order.

#     Raises
#     ------
#     NetworkXError
#         Topological sort is defined for directed graphs only. If the graph `G`
#         is undirected, a :exc:`NetworkXError` is raised.

#     NetworkXUnfeasible
#         If `G` is not a directed acyclic graph (DAG) no topological sort exists
#         and a :exc:`NetworkXUnfeasible` exception is raised.  This can also be
#         raised if `G` is changed while the returned iterator is being processed

#     RuntimeError
#         If `G` is changed while the returned iterator is being processed.

#     Notes
#     -----
#     This algorithm is based on a description and proof in
#     "Introduction to Algorithms: A Creative Approach" [1]_ .

#     See also
#     --------
#     topological_sort

#     References
#     ----------
#     .. [1] Manber, U. (1989).
#        *Introduction to Algorithms - A Creative Approach.* Addison-Wesley.
#     """
#     if not G.is_directed():
#         msg = "Topological sort not defined on undirected graphs."
#         raise nx.NetworkXError(msg)

#     if key is None:

#         def key(node):
#             return node

#     nodeid_map = {n: i for i, n in enumerate(G)}

#     def create_tuple(node):
#         return key(node), nodeid_map[node], node

#     indegree_map = {v: d for v, d in G.in_degree() if d > 0}
#     # These nodes have zero indegree and ready to be returned.
#     zero_indegree = [create_tuple(v) for v, d in G.in_degree() if d == 0]
#     heapq.heapify(zero_indegree)

#     while zero_indegree:
#         _, _, node = heapq.heappop(zero_indegree)

#         if node not in G:
#             raise RuntimeError("Graph changed during iteration")
#         for _, child in G.edges(node):
#             try:
#                 indegree_map[child] -= 1
#             except KeyError as e:
#                 raise RuntimeError("Graph changed during iteration") from e
#             if indegree_map[child] == 0:
#                 heapq.heappush(zero_indegree, create_tuple(child))
#                 del indegree_map[child]

#         yield node

#     if indegree_map:
#         msg = "Graph contains a cycle or graph changed during iteration"
#         raise nx.NetworkXUnfeasible(msg)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any, Dict
from typing import Generator as GenType
from typing import Mapping, Optional, Sequence, TypeVar, Union

from pydantic import PrivateAttr

from dbgen.core.node.computational_node import ComputationalNode

extractor_type = GenType[Dict[str, Mapping[str, Any]], None, None]

T = TypeVar('T')
T1 = TypeVar('T1')


class Extract(ComputationalNode[T]):
    """
    Base Class for all extraction steps.

    Must implement the extract method which returns a list of rows to map the transforms over.

    Must also be subscriptable to connect to the transforms when coding a model.
    """

    _extractor: GenType[Union[Dict[str, Any], T], None, None] = PrivateAttr(None)

    class Config:
        """Pydantic Config"""

        underscore_attrs_are_private = True

    # Overwrite these when writing custom extractor
    def setup(self, **_) -> Optional[GenType[Any, None, None]]:
        pass

    def extract(self: 'Extract[T]') -> GenType[T, None, None]:
        yield {}  # type: ignore

    def teardown(self, **_):
        pass

    def length(self, **_):
        return None

    # Internal Do not Overwrite

    def run(self, _: Dict[str, Mapping[str, Any]]) -> Optional[Dict[str, Any]]:
        try:
            output = next(self._extractor)
            if output is None:
                return {}
            elif isinstance(output, Mapping):
                return dict(output)
            elif isinstance(output, Sequence):
                l1, l2 = len(output), len(self.outputs)
                if l1 != l2:
                    raise ValueError(f"Expected {l2} from extract {self}, but got {l1}")
                return {name: val for name, val in zip(self.outputs, output)}
            else:
                if len(self.outputs) != 1:
                    raise ValueError(
                        f"{self} expected multiple outputs but output a {type(output)} which cannot have its length measured"
                    )
                return {list(self.outputs)[0]: output}
        except StopIteration:
            return None

    def set_extractor(self, *, connection=None, yield_per=None):
        extractor = self.setup(connection=connection, yield_per=yield_per)
        # Check if setup assigned the extractor already
        if extractor is not None:
            self._extractor = extractor
        elif getattr(self, 'extract', None) and callable(self.extract):
            self._extractor = self.extract()
        else:
            raise NotImplementedError(f"No way of getting the extractor from extract {self}")

    def __str__(self):
        return f"{self.__class__.__qualname__}<outputs= {self.outputs}>"



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from pprint import pformat
from typing import Set, Tuple

from pydantic import Field

from dbgen.core.base import Base

set_field = Field(default_factory=lambda: set())


class Dependency(Base):
    tables_needed: Set[str] = set_field
    columns_needed: Set[str] = set_field
    tables_yielded: Set[str] = set_field
    columns_yielded: Set[str] = set_field

    def all(self) -> Tuple[str, str, str, str]:
        a, b, c, d = tuple(
            map(
                lambda x: ",".join(sorted(x)),
                [
                    self.tables_needed,
                    self.columns_needed,
                    self.tables_yielded,
                    self.columns_yielded,
                ],
            )
        )
        return a, b, c, d

    def __str__(self) -> str:
        return pformat(self.__dict__)

    # Public Methods #

    def test(self, other: "Dependency") -> bool:
        """Test whether SELF depends on OTHER"""
        return not (
            self.tables_needed.isdisjoint(other.tables_yielded)
            and self.columns_needed.isdisjoint(other.columns_yielded)
        )

    def merge(self, other: "Dependency") -> "Dependency":
        """Combine two dependencies"""
        return Dependency(
            tables_needed=self.tables_needed.union(other.tables_needed),
            columns_needed=self.columns_needed.union(other.columns_needed),
            tables_yielded=self.tables_yielded.union(other.tables_yielded),
            columns_yielded=self.columns_yielded.union(other.columns_yielded),
        )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import TYPE_CHECKING, Any, Dict
from typing import Generator as GenType
from typing import Optional, TypeVar, Union, overload

from pydantic import Field
from sqlalchemy import text
from sqlmodel.engine.result import Result
from sqlmodel.sql.expression import Select, SelectOfScalar

from dbgen.core.dependency import Dependency
from dbgen.core.node.extract import Extract
from dbgen.core.statement_parsing import _get_select_keys, get_statement_dependency
from dbgen.utils.sql import Connection

if TYPE_CHECKING:
    from sqlalchemy.engine import Connection as SAConnection  # pragma: no cover
    from sqlmodel import Session  # pragma: no cover


SCHEMA_DEFAULT = "public"

T = TypeVar('T')


class BaseQuery(Extract[T]):
    query: str
    params: Dict[str, Any] = Field(default_factory=dict)
    dependency: Dependency = Field(default_factory=Dependency)

    def _get_dependency(self) -> Dependency:
        return self.dependency

    @classmethod
    def from_select_statement(
        cls, select_statement: Select[T], connection: Connection = None, **kwargs
    ) -> "BaseQuery[T]":
        columns, tables, fks = get_statement_dependency(select_statement)
        outputs = _get_select_keys(select_statement)
        get_table_name = lambda x: f"{x.schema}.{x.name}" if getattr(x, "schema", "") else x.name
        dependency = Dependency(
            tables_needed=[get_table_name(x) for x in tables],
            columns_needed=[f"{get_table_name(x.table)}.{x.name}" for x in columns.union(fks)],
        )
        compiled_statement = select_statement.compile()
        return cls(
            inputs=[],
            outputs=outputs,
            query=str(compiled_statement),
            params=compiled_statement.params,
            dependency=dependency,
        )

    def render_query(self) -> str:
        """Stringifies the query with the bound parameters."""
        compiled_query = (
            text(self.query).bindparams(**self.params).compile(compile_kwargs={'literal_binds': True})
        )
        return str(compiled_query)

    def length(self, *, connection: 'SAConnection' = None, **_) -> int:
        assert connection
        count_statement = f"select count(1) from ({self.query}) as X;"
        rows: int = connection.execute(text(count_statement), **self.params).scalar()  # type: ignore
        return rows

    def setup(
        self,
        connection: 'SAConnection' = None,
        yield_per: Optional[int] = None,
        **kwargs,
    ) -> GenType[Result[T], None, None]:
        assert connection, f"Need to pass in connection when setting the extractor"
        if yield_per:
            result = connection.execution_options(stream_results=True).execute(
                text(self.query), **self.params
            )
            while chunk := result.fetchmany(yield_per):
                for row in chunk:
                    yield dict(row)  # type: ignore
        else:
            result = connection.execute(text(self.query), **self.params)
            yield from result.mappings()  # type: ignore


class ExternalQuery(BaseQuery[T]):
    connection: Connection

    @classmethod
    def from_select_statement(
        cls, select_statement: Select, connection: 'Connection' = None, **kwargs
    ) -> "ExternalQuery":
        selected_keys = _get_select_keys(select_statement)
        return cls(
            query=str(select_statement),
            outputs=selected_keys,
            connection=connection,
        )

    def setup(
        self,
        connection: Union['SAConnection', 'Session'] = None,
        yield_per: Optional[int] = None,
        **kwargs,
    ) -> GenType[Result[T], None, None]:
        engine = self.connection.get_engine()
        with engine.connect() as conn:
            if yield_per:
                result = conn.execute(text(self.query), **self.params)
                yield from result.yield_per(yield_per).mappings()  # type: ignore
            else:
                result = conn.execute(text(self.query), **self.params)
                yield from result.mappings()  # type: ignore


@overload
def Query(select_statement: SelectOfScalar[T], connection: Connection) -> ExternalQuery[T]:
    ...


@overload
def Query(select_statement: SelectOfScalar[T], connection: None = None) -> BaseQuery[T]:
    ...


@overload
def Query(select_statement: Select[T], connection: Connection) -> ExternalQuery[T]:
    ...


@overload
def Query(select_statement: Select[T], connection: None = None) -> BaseQuery[T]:
    ...


def Query(select_statement, connection: Optional[Connection] = None) -> Union[BaseQuery, ExternalQuery]:
    cls = BaseQuery if connection is None else ExternalQuery
    return cls.from_select_statement(select_statement, connection=connection)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import re
from pathlib import Path
from typing import Generator as GenType
from typing import List, Optional

from pydantic import DirectoryPath

from dbgen.core.node.extract import Extract


class FileExtractor(Extract[Path]):
    directory: DirectoryPath
    pattern: Optional[str]
    recursive: bool = False
    outputs: List[str] = ['file_name']
    _file_paths: List[Path]

    def setup(self, **_):
        file_gen = self.directory.rglob('*') if self.recursive else self.directory.iterdir()
        self._file_paths = [
            x for x in file_gen if x.is_file() and (self.pattern is None or re.search(self.pattern, x.name))
        ]

    def extract(self) -> GenType[Path, None, None]:
        yield from self._file_paths

    def length(self, **_):
        return len(self._file_paths)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import contextlib
import logging
from enum import Enum
from io import StringIO
from logging import Formatter

from rich.console import Console
from rich.logging import RichHandler
from rich.traceback import install

install()
console = Console()


class LogLevel(str, Enum):
    DEBUG = 'DEBUG'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'
    CRITICAL = 'CRITICAL'


def get_log_level(log_level: LogLevel):
    return getattr(logging, log_level)


def capture_stdout(func):
    def wrapped(*args, **kwargs):
        logger = logging.getLogger(f"dbgen.pyblock.{func.name}")
        stream = StringIO()
        with contextlib.redirect_stdout(stream):
            output = func(*args, **kwargs)
        stdout = stream.getvalue().strip()
        if stdout:
            logger.debug(stdout)
        return output

    return wrapped


def setup_logger(level: LogLevel = LogLevel.DEBUG, log_to_stdout=True):
    custom_logger = logging.getLogger("dbgen")
    custom_logger.propagate = True
    custom_logger.setLevel(get_log_level(level))
    return custom_logger


def add_stdout_logger(logger, stdout_level: LogLevel = LogLevel.DEBUG):
    level_val = get_log_level(stdout_level)
    rich_handler = RichHandler(level=level_val, markup=True, console=console)
    log_format = r"[magenta]\[%(name)s][/magenta] - %(message)s"
    rich_handler.setFormatter(Formatter(log_format))
    logger.addHandler(rich_handler)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from datetime import datetime
from logging import getLogger
from pathlib import Path
from typing import List, Optional
from uuid import UUID

import typer
from rich.console import Console
from rich.table import Column, Table
from rich.text import Text
from sqlmodel import Session, select

import dbgen.cli.styles as styles
import dbgen.exceptions as exceptions
from dbgen.cli.options import config_option, model_string_option, version_option
from dbgen.cli.queries import get_runs
from dbgen.cli.utils import confirm_nuke, set_confirm, test_connection, validate_model_str
from dbgen.configuration import config, initialize, root_logger
from dbgen.core.metadata import GeneratorRunEntity, ModelEntity, RunEntity
from dbgen.core.run import RunConfig
from dbgen.utils.log import LogLevel, add_stdout_logger

run_app = typer.Typer(name='run')
logger = getLogger(__name__)


@run_app.command('status')
def status(
    run_id: Optional[int] = None,
    last: int = 1,
    all_runs: bool = typer.Option(False, '-a', '--all', help='show all runs ever'),
    error: bool = typer.Option(False, '--error', help='show error for failed runs'),
    query: bool = typer.Option(False, '--query', help='show error for failed runs'),
    statuses: List[str] = typer.Option(
        [], '-s', '--status', help='Filter out only statuses that match input val'
    ),
):
    column_dict = {'error': Column('Error'), 'status': Column('Status', style='green')}
    _, meta_conn = initialize()
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    filtered_keys = {'generator_id', 'created_at'}
    if not error:
        filtered_keys.add('error')
    if not query:
        filtered_keys.add('query')
    columns = [
        'name',
        *(c.name for c in GeneratorRunEntity.__table__.c if c.name not in filtered_keys),
    ]
    show_cols = {col: col.replace('number_of_', '') for col in columns}
    table = Table(
        *[column_dict.get(col, col) for col in show_cols],
        title='run',
        show_lines=True,
        highlight=True,
        border_style='magenta',
    )
    status_map = {'failed': 'white on red', 'completed': 'green', 'excluded': 'grey'}

    def style(key, col):
        if key == 'status':
            return Text(str(col), style=status_map.get(col))
        elif key == 'memory_usage':
            return f"{col:3.1f} MB" if col else ''
        elif key == 'runtime':
            return f"{col} (s)" if col else 'N/A'
        return str(col)

    for row in get_runs(run_id, meta_engine, all_runs, statuses, last):
        table.add_row(
            *[style(key, row[key]) for (key, col) in show_cols.items()],
        )
    console = Console()
    console.print(table)


@run_app.callback(invoke_without_command=True)
def run_model(
    ctx: typer.Context,
    model_str: str = model_string_option,
    include: List[str] = typer.Option([], help="Generators to include"),
    exclude: List[str] = typer.Option([], help="Generators to xclude"),
    retry: bool = typer.Option(False, help="Ignore repeat checking"),
    start: Optional[str] = typer.Option(None, help="Generator to start run at"),
    until: Optional[str] = typer.Option(None, help="Generator to finish run at."),
    level: LogLevel = typer.Option(LogLevel.INFO, help='Use the RemoteGenerator Runner'),
    bar: bool = typer.Option(True, help="Show progress bar"),
    skip_row_count: bool = typer.Option(False, help="Show progress bar"),
    skip_on_error: bool = typer.Option(False, help="Skip a row in generator on error"),
    batch: Optional[int] = typer.Option(None, help="Batch size for all generators in run."),
    batch_number: int = typer.Option(10, help="Default number of batches per generator."),
    pdb: bool = typer.Option(False, '--pdb', help="Drop into pdb on breakpoints"),
    config_file: Path = config_option,
    remote: bool = typer.Option(True, help='Use the RemoteGenerator Runner'),
    no_conf: bool = typer.Option(
        False,
        "--no-confirm",
        "-y",
        is_eager=True,
        callback=set_confirm,
    ),
    rerun_failed: bool = typer.Option(
        False,
        help="Only rerun the generators that failed or were excluded in the previous run.",
    ),
    nuke: bool = typer.Option(
        None,
        help="Delete the entire db and meta schema.",
        callback=confirm_nuke,
    ),
    version: bool = version_option,
):
    """Run a model."""
    if ctx.invoked_subcommand is not None:
        return
    # Start connection from config
    main_conn, meta_conn = initialize(config_file)
    # Use config model_str if none is provided
    if model_str is None:
        model_str = config.model_str
    # validate the model_str
    model = validate_model_str(model_str)
    styles.delimiter(styles.typer.colors.GREEN)
    styles.good_typer_print(f"Running model {model.name}...")
    styles.delimiter(styles.typer.colors.GREEN)
    run_config = RunConfig(
        retry=retry,
        start=start,
        until=until,
        exclude=exclude,
        include=include,
        progress_bar=bar,
        log_level=level,
        skip_row_count=skip_row_count,
        skip_on_error=skip_on_error,
        batch_size=batch,
        batch_number=batch_number,
    )
    if not bar:
        add_stdout_logger(root_logger, stdout_level=run_config.log_level)
    # Validate the run configuration parameters
    invalid_run_config_vals = run_config.get_invalid_markers(model)
    if invalid_run_config_vals:
        lines = [f"{name}: {' ,'.join(val)}" for name, val in invalid_run_config_vals.items()]
        error = "\n".join(lines)
        raise typer.BadParameter(f"Invalid run configuration parameters passed:\n{error}")

    # Test each connection with simple select command
    logger.debug('testing connections')
    for name, conn in (('main', main_conn), ('meta', meta_conn)):
        logger.debug(f'testing {name} connection...')
        test_connection(conn, name)
    # Grab engine from each connection
    main_engine, meta_engine = main_conn.get_engine(), meta_conn.get_engine()
    # Pass all the arguments to the model run command
    if pdb:
        config.pdb = pdb
    try:
        out_run = model.run(
            main_engine, meta_engine, run_config, nuke=nuke, rerun_failed=rerun_failed, remote=remote
        )
    except exceptions.SerializationError as exc:
        raise typer.BadParameter(
            f"A generator in your model could not be deserialized. Did you use a custom extractor?"
        ) from exc

    # Once run is done talk to the meta database to report to the user how the run went
    with Session(meta_engine) as session:
        run = session.exec(select(RunEntity).where(RunEntity.id == out_run.id)).one()
        failed_gens = [gen_run.generator.name for gen_run in run.generator_runs if gen_run.status == 'failed']
        rows_inserted = sum([gen_run.rows_inserted for gen_run in run.generator_runs])
        rows_updated = sum([gen_run.rows_updated for gen_run in run.generator_runs])

    if run.errors:
        styles.delimiter(styles.typer.colors.RED)
        styles.bad_typer_print(f"Encountered {run.errors} during the run! The following generators failed:")
        styles.bad_typer_print('\n'.join(map(repr, failed_gens)))
    if run.runtime:
        styles.delimiter(styles.typer.colors.GREEN)
        styles.good_typer_print(
            f"Finished Running {len(model.generators)} Generator(s) in {run.runtime.total_seconds():.3f}(s)."
        )
        styles.good_typer_print(
            f"The run attempted to insert {rows_inserted} and update {rows_updated} rows."
        )
    raise typer.Exit(code=0)


@run_app.command('initialize')
def run_initialize(model_id: UUID, config_file: Path = config_option):
    """Initialize a run."""
    # Notify of config file
    if config_file:
        _, meta_conn = initialize(config_file)
    test_connection(meta_conn)
    meta_engine = meta_conn.get_engine()
    typer.echo(config.display())
    with Session(meta_engine) as session:
        model = session.get(ModelEntity, model_id)
        if not model:
            raise ValueError(f"Invalid model id, could not fine model with ID {model_id}")
        model.last_run = datetime.now()
        run = RunEntity(status='initialized')
        session.add(run)
        session.commit()
        session.refresh(run)
    typer.echo(run.id)
    return



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from enum import Enum
from pathlib import Path

import typer
from cookiecutter.main import cookiecutter  # type: ignore

# Create project from the cookiecutter-pypackage.git repo template
new_app = typer.Typer(name='new')


class Template(str, Enum):
    SIMPLE = 'simple'
    COMPLEX = 'complex'


complexity_map = {
    Template.SIMPLE: 'https://github.com/modelyst/dbgen-model-template',
}


@new_app.callback(invoke_without_command=True)
def new(
    template: Template = Template.SIMPLE,
    overwrite_if_exists: bool = typer.Option(False, '--overwrite'),
    output_dir: Path = typer.Option(Path('.'), '--output', '-o'),
    config_file: Path = typer.Option(None, '--config', '-c'),
):
    template_url = complexity_map.get(template, template)
    typer.secho(f'Downloading template from {template_url}...', fg='green')
    cookiecutter(
        template_url, overwrite_if_exists=overwrite_if_exists, output_dir=output_dir, config_file=config_file
    )



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""Utilities for printing things to screen for CLI."""
import typer

from dbgen import __version__

THEME_COLOR = typer.colors.BRIGHT_MAGENTA
LOGO = """
    ____  ____
   / __ \\/ __ )____ ____  ____
  / / / / __  / __ `/ _ \\/ __ \\
 / /_/ / /_/ / /_/ /  __/ / / /
/_____/_____/\\__, /\\___/_/ /_/
            /____/
"""

PRINT_LOGO = f"""
-----------------------------------{LOGO}-----------------------------------
VERSION: {__version__}
-----------------------------------
"""


def delimiter(color: str = THEME_COLOR):
    typer.echo(typer.style("---------------------------------------------", fg=color, bold=True))


LOGO_STYLE = typer.style(PRINT_LOGO, blink=True, fg=THEME_COLOR)
# Easy printers
typer_print = lambda color: lambda msg: typer.echo(typer.style(msg, fg=color))
good_typer_print = typer_print(typer.colors.GREEN)
bad_typer_print = typer_print(typer.colors.RED)
theme_typer_print = typer_print(THEME_COLOR)
greens = lambda x: typer.style(x, fg=typer.colors.BRIGHT_GREEN)
reds = lambda x: typer.style(x, fg=typer.colors.BRIGHT_RED)



#   Copyright 2021 Modelyst LLC
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from collections import Counter
from typing import TYPE_CHECKING, Dict, Iterable, List, Optional, Tuple
from uuid import UUID

import sqlalchemy
from networkx import DiGraph
from pydantic import Field, PrivateAttr, validator
from sqlalchemy import MetaData, inspect, text
from sqlalchemy.future import Engine
from sqlalchemy.orm import registry as sa_registry

from dbgen.core.base import Base
from dbgen.core.context import ModelContext
from dbgen.core.entity import BaseEntity
from dbgen.core.generator import Generator
from dbgen.core.metadata import ModelEntity, RunEntity, meta_registry
from dbgen.exceptions import DatabaseError
from dbgen.utils.graphs import serialize_graph, topsort_with_dict

if TYPE_CHECKING:
    from dbgen.core.run import RunConfig  # pragma: no cover


class Model(Base):
    name: str
    generators: List[Generator] = Field(default_factory=list)
    registry: sa_registry = Field(default_factory=lambda: BaseEntity._sa_registry)
    meta_registry: sa_registry = Field(default_factory=lambda: meta_registry)
    _context: ModelContext = PrivateAttr(None)
    _hashinclude_ = {"name", "generators"}

    class Config:
        """Pydantic Config"""

        arbitrary_types_allowed = True

    def __enter__(self):
        self._context = ModelContext(context_dict={'model': self})
        return self._context.__enter__()['model']

    def __exit__(self, *args):
        self._context.__exit__(*args)
        self.validate(self)
        del self._context

    @validator("generators")
    def unique_generator_names(cls, value: List[Generator]) -> List[Generator]:
        name_count = Counter([x.name for x in value])
        dup_names = {name for name, count in name_count.items() if count > 1}
        if dup_names:
            raise ValueError(f"Found duplicate generator names, all names must be unique: {dup_names}")
        return value

    @validator("generators")
    def sort_generators(cls, value: List[Generator]) -> List[Generator]:
        return value

    def gens(self) -> Dict[str, Generator]:
        return {gen.name: gen for gen in self.generators}

    def add_gen(self, generator: Generator):
        if generator.name not in self.gens():
            self.generators.append(generator)
            return
        raise ValueError(
            f"Generator named {generator.name} already in model please use a new name:\n{self.gens().keys()}"
        )

    def validate_marker(self, marker: str) -> bool:
        """Compare a marker to the models generators to see if it is valid."""
        valid_marker = False
        for gen in self.generators:
            if marker == gen.name or marker in gen.tags:
                valid_marker = True
        return valid_marker

    def _generator_graph(self) -> "DiGraph":
        from networkx import DiGraph

        graph = DiGraph()

        # Add edges for every Arg in graph
        edges: List[Tuple[str, str]] = []
        for source_name, source in self.gens().items():
            for target_name, target in self.gens().items():
                if source_name != target_name:
                    if source._get_dependency().test(target._get_dependency()):
                        edges.append((target_name, source_name))

        for gen in self.gens().values():
            graph.add_node(gen.name, id=gen.uuid, **gen.dict(include={'name', 'dependency'}))
        graph.add_edges_from(edges)
        return graph

    def _sort_graph(self) -> List[Generator]:
        graph = self._generator_graph()
        gens = self.gens()
        sorted_node_ids = topsort_with_dict(graph)
        sorted_nodes = [gens[key] for key in sorted_node_ids]
        return sorted_nodes

    def run(
        self,
        main_engine: Engine,
        meta_engine: Engine,
        run_config: "RunConfig" = None,
        nuke: bool = False,
        rerun_failed: bool = False,
        remote: bool = True,
    ) -> RunEntity:
        from dbgen.core.run import ModelRun, RemoteModelRun

        Runner = RemoteModelRun if remote else ModelRun
        return Runner(model=self).execute(
            main_engine=main_engine,
            meta_engine=meta_engine,
            run_config=run_config,
            nuke=nuke,
            rerun_failed=rerun_failed,
        )

    def sync(self, main_engine: Engine, meta_engine: Engine, nuke: bool = False) -> None:
        """Syncs the state of the models registry with the database."""
        # Inspect the schemas and make sure they all exist
        for engine, metadata in (
            (main_engine, self.registry.metadata),
            (meta_engine, self.meta_registry.metadata),
        ):
            inspector = inspect(engine)
            expected_schemas = {x.schema for x in metadata.tables.values() if x.schema}
            current_schemas = inspector.get_schema_names()

            # Nuking drops all tables in the schemas of the model
            if nuke:
                self.nuke(engine, metadata=metadata, schemas=expected_schemas)

            missing_schema = {x for x in expected_schemas if x not in current_schemas}
            for schema in missing_schema:
                self._logger.debug(f"Missing schema {schema}, creating now")
                with engine.begin() as conn:
                    conn.execute(text(f"CREATE SCHEMA {schema}"))

        # Create all tables
        self.registry.metadata.create_all(main_engine)

        # Create meta schema
        self.meta_registry.metadata.create_all(meta_engine)

    def nuke(
        self,
        engine: Engine,
        metadata: MetaData,
        schemas: Optional[Iterable[Optional[str]]] = None,
        nuke_all: bool = False,
    ) -> None:
        # Validate and set the schema
        schemas = schemas or []
        if not schemas:
            # Nuke all grabs all schemas in the database
            # Otherwise we use None and the default schema in the engine will be dropped
            if nuke_all:
                schemas = inspect(engine).get_schema_names()
            else:
                schemas = [None]

        assert isinstance(schemas, Iterable), f"schemas must be iterable: {schemas}"
        # Drop the expected tables
        try:
            metadata.drop_all(engine)
        except sqlalchemy.exc.InternalError as exc:
            raise DatabaseError("Error occurred during nuking. Please drop the schema manually...") from exc
        # Iterate through the schemas this metadata describes and drop all the tables within them
        for schema in schemas:
            self._logger.info(f"Nuking the schema={schema!r}...")
            metadata = MetaData()
            metadata.reflect(engine, schema=schema)
            try:
                metadata.drop_all(engine)
            except sqlalchemy.exc.InternalError:
                with engine.begin() as conn:
                    conn.execute(text(f'DROP SCHEMA {schema} CASCADE'))
                # raise ValueError(
                #     f"Nuking has failed! Dependent objects exist that SQL alchemy cannot successfully drop from the schema {schema!r}.\n"
                #     f"This often occurs when Views exist in the schema. Please drop these views manually by running 'DROP SCHEMA {schema} CASCADE'"
                # )
            self._logger.info(f"Schema {schema!r} nuked.")

    def _get_model_row(self):
        graph = self._generator_graph()

        def default(obj):
            if isinstance(obj, set):
                return list(obj)
            elif isinstance(obj, UUID):
                return str(obj)
            raise TypeError(f"Unknown type: {type(obj)}")

        metadata = {'name': self.name, 'id': self.uuid}
        graph_json = serialize_graph(graph, lambda x: x, metadata=metadata)
        # Assemble stringified dependency fields as we can't store sets in postgres easily
        return ModelEntity(
            id=self.uuid,
            name=self.name,
            generators=[x._get_gen_row() for x in self.generators],
            graph_json=graph_json,
        )
